Require Import Bool.
Require Import List. 
Require Import finset.

Record Mealy (Σ : Alphabet) (Γ : Alphabet) :=
{
   Mealy_st : FinType;
   Mealy_init : Mealy_st;
   Mealy_delta : Σ * Mealy_st -> Γ * Mealy_st;
}.

Arguments Mealy_st [_ _] _.
Arguments Mealy_init [_ _ _].
Arguments Mealy_delta [_ _ _] _.

Fixpoint run_of_Mealy {Σ Γ : Alphabet} (M : Mealy Σ Γ) (u : nat -> Σ) (n : nat) : Γ * (Mealy_st M) :=
  match n with
   0 => Mealy_delta (u 0, Mealy_init)
 | S k => let (_,q) := run_of_Mealy M u k in
          Mealy_delta (u (S k), q)
  end.

Definition fun_of_Mealy {Σ Γ} (M : Mealy Σ Γ) : (nat -> Σ) -> (nat -> Γ) :=
  fun u n => fst (run_of_Mealy M u n).

Definition fun0_of_Mealy {Σ Γ} (M : Mealy Σ Γ) : Σ -> Γ :=
  fun a => fun_of_Mealy M (fun _ => a) 0.

Lemma fun0_of_Mealy_eq {Σ Γ} (M : Mealy Σ Γ) : forall a : Σ,
   fun0_of_Mealy M a = fst (@Mealy_delta _ _ M (a, Mealy_init)).
  intros; reflexivity.
Qed.

Definition next_Mealy {Σ Γ : Alphabet} (a₀ : Σ) (M : Mealy Σ Γ) : Mealy Σ Γ :=
  {|
     Mealy_st := Mealy_st M;
     Mealy_init := snd (Mealy_delta (a₀, Mealy_init));
     Mealy_delta := @Mealy_delta _ _ M;
  |}.

Fixpoint suff_Mealy {Σ Γ : Alphabet} (p : list Σ) (M : Mealy Σ Γ) : Mealy Σ Γ :=
  match p with
   nil => M
 | a :: p' => next_Mealy a (suff_Mealy p' M)
  end.

Definition Mealy_memoryless {Σ Γ : Alphabet} (f : Σ -> Γ) : Mealy Σ Γ :=
 {|
    Mealy_st := Fin_unit;
    Mealy_init := tt;
    Mealy_delta := fun z => let (a, tt) := z in
                     (f a, tt);
 |}.

Definition Mealy_fix {Σ Γ : Alphabet} (b₀ : Γ) (M : Mealy (Al_prod Σ Γ) Γ) : Mealy Σ Γ :=
 {|
    Mealy_st := Fin_prod (Mealy_st M) Γ;
    Mealy_init := (Mealy_init, b₀);
    Mealy_delta := fun z =>
                    let '(a, (q, b)) := z in
                    let (b', q') := @Mealy_delta _ _ M ((a,b),q) in
                    (b', (q',b'));
 |}.

Definition Mealy_comp {Σ Γ Θ : Alphabet} (N : Mealy Γ Θ) (M : Mealy Σ Γ) : Mealy Σ Θ :=
 {|
    Mealy_st := Fin_prod (Mealy_st M) (Mealy_st N);
    Mealy_init := (Mealy_init, Mealy_init);
    Mealy_delta := fun z =>
                   let '(a,(qM, qN)) := z in
                   let (b, qM') := Mealy_delta (a, qM) in
                   let (c, qN') := Mealy_delta (b, qN) in
                   (c, (qM',qN'));
 |}.

Definition Mealy_cons {Σ : Alphabet} (a₀ : Σ) : Mealy Σ Σ :=
 Mealy_comp (@Mealy_memoryless (Al_prod Σ Σ) _ snd)
             (@Mealy_fix Σ (Al_prod Σ Σ) (a₀,a₀)
                 (@Mealy_memoryless (Al_prod Σ (Al_prod Σ Σ)) (Al_prod Σ Σ)
                   (fun z => let '(a, (b,_)) := z in (a,b)))).

Lemma Mealy_comp_run_eq : forall (Σ Γ Θ : Alphabet) (N : Mealy Γ Θ) (M : Mealy Σ Γ) u,
  forall n, let '(c, (qM, qN)) := run_of_Mealy (Mealy_comp N M) u n in
            let (_, qM') := run_of_Mealy M u n in
            let (c', qN') := run_of_Mealy N (fun_of_Mealy M u) n in
            (qM, qN, c) = (qM', qN', c').
  unfold fun_of_Mealy; induction n; simpl.
  + destruct (Mealy_delta (u 0, Mealy_init)); simpl.
    destruct (Mealy_delta (f, Mealy_init)); reflexivity.
  + destruct (run_of_Mealy (Mealy_comp N M) u n).
    destruct f0, (run_of_Mealy M u n); subst.
    remember ((run_of_Mealy N (fun n => fst (run_of_Mealy M u n))) n) as α.
    destruct α.
    injection IHn; intros; subst; clear IHn.
    destruct (Mealy_delta (u (S n), f3)); simpl.
    destruct (Mealy_delta (f, f5)); reflexivity.
Qed.

Lemma Mealy_comp_fun_eq : forall (Σ Γ Θ : Alphabet) (N : Mealy Γ Θ) (M : Mealy Σ Γ) u,
  fun_of_Mealy (Mealy_comp N M) u == fun_of_Mealy N (fun_of_Mealy M u).
  intros;
  specialize (Mealy_comp_run_eq Σ Γ Θ N M u x); intros;
  unfold fun_of_Mealy in *.
  destruct (run_of_Mealy (Mealy_comp N M) u x),
           (run_of_Mealy M u x),
           (run_of_Mealy N (fun n => fst (run_of_Mealy M u n)) x).
  destruct f0; injection H; simpl; auto.
Qed.

Lemma Mealy_memoryless_fun_eq : forall (Σ Γ : Alphabet) (f : Σ -> Γ) u,
  fun_of_Mealy (Mealy_memoryless f) u == (fun k => f (u k)).
  intros.
  destruct x;[reflexivity|].
  unfold fun_of_Mealy; simpl.
  destruct (run_of_Mealy (Mealy_memoryless f)); reflexivity.
Qed.

Definition str_cons {Σ : Type} (a₀ : Σ) (u : nat -> Σ) (n : nat) : Σ :=
  match n with
   0 => a₀
 | S k => u k
  end.

Lemma Mealy_cons_run_eq : forall (Σ : Alphabet) (a₀ : Σ) u,
  forall n,
   let '(a, ((tt, (a', a'')), tt)) := run_of_Mealy (Mealy_cons a₀) u (S n) in
   (a, a', a'') = (u n, u (S n), u n).
  induction n; [reflexivity|].
  simpl in *.
  destruct (run_of_Mealy (Mealy_cons a₀) u n).
  destruct f0 as [[[] [a b]] []]; reflexivity.
(* goes through but lags a lot (30s) *)
Admitted.

Lemma Mealy_cons_fun_eq : forall (Σ : Alphabet) (a₀ : Σ) u,
  fun_of_Mealy (Mealy_cons a₀) u == str_cons a₀ u.
  intros ? ? ? [|n]; [reflexivity |specialize (Mealy_cons_run_eq Σ a₀ u n)];
  unfold fun_of_Mealy.
  destruct (run_of_Mealy (Mealy_cons a₀) u (S n)); simpl.
  destruct f0 as [[] ? ].
  destruct f1, f0, f2. 
  intro H; injection H; auto.
Qed.

Lemma Mealy_fix_run_eq : forall (Σ Γ : Alphabet) (M : Mealy (Al_prod Σ Γ) Γ) (b₀ : Γ) u,
  forall n,
  let consfix := str_cons b₀ (fun_of_Mealy (Mealy_fix b₀ M) u) in
  let '(b, (q, bq)) := run_of_Mealy (Mealy_fix b₀ M) u n in
  let '(b', q') := run_of_Mealy M (pairfun u consfix) n in
  (b, q, bq) = (b', q', b).
  unfold pairfun; induction n; unfold fun_of_Mealy in *.
  + simpl.
    destruct (@Mealy_delta _ _ M (u 0, b₀, Mealy_init)); reflexivity.
  + simpl in *.
    destruct (run_of_Mealy (Mealy_fix b₀ M) u n).
    destruct f0. simpl in *.
    destruct (@run_of_Mealy (Al_prod Σ Γ) Γ M
             (fun k : nat =>
              (u k,
              @str_cons Γ b₀
                (fun n : nat =>
                 @fst Γ (@Mealy_st (Al_prod Σ Γ) Γ M * Γ)
                   (@run_of_Mealy Σ Γ (@Mealy_fix Σ Γ b₀ M) u n)) k)) n).
    injection IHn; intros; subst; clear IHn.
    destruct (@Mealy_delta _ _ M (u (S n), f2, f3)); reflexivity.
Qed.

Lemma Mealy_fix_fun_eq : forall (Σ Γ : Alphabet) (M : Mealy (Al_prod Σ Γ) Γ) (b₀ : Γ) u,
  fun_of_Mealy (Mealy_fix b₀ M) u == fun_of_Mealy M (pairfun u (str_cons b₀ (fun_of_Mealy (Mealy_fix b₀ M) u))).
  unfold fun_of_Mealy.
  intros.
  specialize (Mealy_fix_run_eq _ _ M b₀ u x).
  simpl; intro.
  destruct (run_of_Mealy (Mealy_fix b₀ M)u x).
  destruct f0; simpl in *.
  replace (@run_of_Mealy (Al_prod Σ Γ) Γ M
       (@pairfun nat Σ Γ u
          (@str_cons Γ b₀
             (fun n : nat =>
              @fst Γ (@Mealy_st (Al_prod Σ Γ) Γ M * Γ)
                (@run_of_Mealy Σ Γ (@Mealy_fix Σ Γ b₀ M) u n)))) x)
  with (@run_of_Mealy (Al_prod Σ Γ) Γ M
         (@pairfun nat Σ Γ u (@str_cons Γ b₀ (@fun_of_Mealy Σ Γ (@Mealy_fix Σ Γ b₀ M) u)))
         x);[|reflexivity].
  destruct (run_of_Mealy M (pairfun u (str_cons b₀ (fun_of_Mealy (Mealy_fix b₀ M) u))) x).
  injection H; auto.
Qed.
 

Record Mealy_morph {Σ Γ : Alphabet} (M M' : Mealy Σ Γ) :=
  {
     Mealy_m_st  : Mealy_st M -> Mealy_st M';
     Mealy_m_init : Mealy_m_st Mealy_init = Mealy_init;
     Mealy_m_delta : forall a q, Mealy_delta (a, Mealy_m_st q) =
                                 let (a', q') := Mealy_delta (a, q) in
                                 (a', Mealy_m_st q');
  }.

Arguments Mealy_m_st [_ _ _ _] _.

Lemma Mealy_morph_run_eq {Σ Γ : Alphabet} :
  forall M N : Mealy Σ Γ, forall F : Mealy_morph M N,
  forall u, (fun n => let (b, q) := run_of_Mealy M u n in
                      (b, Mealy_m_st F q)) == run_of_Mealy N u.
  intros.
  induction x; simpl.
  - rewrite <-Mealy_m_delta, Mealy_m_init; reflexivity.
  - rewrite <-IHx; destruct (run_of_Mealy M u x).
    rewrite Mealy_m_delta; reflexivity.
Qed.

Lemma Mealy_morph_fun_eq {Σ Γ : Alphabet} :
  forall M N : Mealy Σ Γ, forall F : Mealy_morph M N,
  forall u, fun_of_Mealy M u == fun_of_Mealy N u.
  intros; unfold fun_of_Mealy; rewrite <-(Mealy_morph_run_eq _ _ F).
  destruct (run_of_Mealy M u x); reflexivity.
Qed.

Definition next_Mealy_fix_morph : forall (Σ Γ : Alphabet) (M : Mealy (Al_prod Σ Γ) Γ) a₀ b₀,
  Mealy_morph (next_Mealy a₀ (Mealy_fix b₀ M))
              (Mealy_fix (fun0_of_Mealy M (a₀, b₀)) (@next_Mealy (Al_prod _ _) _ (a₀, b₀) M)).
  intros.
  unshelve esplit; [exact (fun x => x) | |].
  - simpl. rewrite fun0_of_Mealy_eq.
    case_eq (@Mealy_delta _ _ M (a₀, b₀, Mealy_init)).
    intros; simpl. rewrite H; reflexivity.
  - intros; simpl.
    destruct q.
    destruct (@Mealy_delta _ _ M (a, f0 , f)); reflexivity.
Defined.

Lemma next_Mealy_fix_eq : forall (Σ Γ : Alphabet) (M : Mealy (Al_prod Σ Γ) Γ) a₀ b₀ u,
  fun_of_Mealy (next_Mealy a₀ (Mealy_fix b₀ M)) u ==
  fun_of_Mealy (Mealy_fix (fun0_of_Mealy M (a₀, b₀)) (@next_Mealy (Al_prod _ _) _ (a₀, b₀) M)) u.
  intros; apply Mealy_morph_fun_eq, next_Mealy_fix_morph.
Qed.

Lemma next_Mealy_run_eq : forall (Σ Γ : Alphabet) (M : Mealy Σ Γ),
  forall u n, run_of_Mealy M u (S n) = run_of_Mealy (next_Mealy (u 0) M) (fun k => u (S k)) n.
  intros; induction n; simpl in *.
  + destruct (Mealy_delta (u 0, Mealy_init)); reflexivity.
  + destruct (run_of_Mealy M u n).
    rewrite <-IHn; reflexivity.
Qed.

Lemma next_Mealy_fun_eq : forall (Σ Γ : Alphabet) (M : Mealy Σ Γ),
  forall u n, fun_of_Mealy M u (S n) = fun_of_Mealy (next_Mealy (u 0) M) (fun k => u (S k)) n.
  intros; unfold fun_of_Mealy; rewrite next_Mealy_run_eq; reflexivity.
Qed.

Lemma run_of_Mealy_ext : forall (Σ Γ : Alphabet) (M : Mealy Σ Γ),
  forall u u', u == u' -> run_of_Mealy M u == run_of_Mealy M u'.
  intros. induction x; simpl; rewrite H; [|rewrite IHx]; reflexivity.
Qed.

Lemma fun_of_Mealy_ext : forall (Σ Γ : Alphabet) (M : Mealy Σ Γ),
  forall u u', u == u' -> fun_of_Mealy M u == fun_of_Mealy M u'.
  intros; unfold fun_of_Mealy; erewrite run_of_Mealy_ext; auto.
Qed.

Lemma Mealy_fix_fun_uniq : forall {Σ Γ : Alphabet} (M : Mealy (Al_prod Σ Γ) Γ) (b₀ : Γ) u v,
  v == fun_of_Mealy M (pairfun u (str_cons b₀ v)) -> v == fun_of_Mealy (Mealy_fix b₀ M) u.
  intros.  
  revert M b₀ u v H.
  induction x.
  - intros.
    rewrite H, Mealy_fix_fun_eq; reflexivity.
  - intros.
    rewrite next_Mealy_fun_eq, next_Mealy_fix_eq.
    apply (IHx (@next_Mealy (Al_prod _ _) _ (u 0, b₀) M) (fun0_of_Mealy M (u 0, b₀)) (fun k => u (S k)) (fun k => v (S k))).
    intros.
    rewrite H, next_Mealy_fun_eq.
    specialize (H 0).
    unfold pairfun in *; simpl in *.
    apply fun_of_Mealy_ext.
    intros; apply f_equal.
    destruct x1; simpl; auto.
Qed.
