Require Import finset.
Require Import mealy.


Inductive tm (Σ Γ : Alphabet) : Type :=
  tm_memoryless : forall f : Σ -> Γ, tm Σ Γ
| tm_comp : forall Θ, tm Θ Γ -> tm Σ Θ -> tm Σ Γ
| tm_fix : forall b₀ : Γ, tm (Al_prod Σ Γ) Γ -> tm Σ Γ.

Arguments tm_memoryless [_ _] _.
Arguments tm_comp [_ _ _] _.
Arguments tm_fix [_ _] _.

Fixpoint Mealy_of_tm {Σ Γ} (t : tm Σ Γ) : Mealy Σ Γ :=
  match t with
  tm_memoryless f => Mealy_memoryless f
| tm_comp N M => Mealy_comp (Mealy_of_tm N) (Mealy_of_tm M)
| tm_fix b₀ M => Mealy_fix b₀ (Mealy_of_tm M)
  end.

Definition fun_of_tm {Σ Γ} (t : tm Σ Γ) := fun_of_Mealy (Mealy_of_tm t).

Definition tm_run_of_Mealy {Σ Γ} (M : Mealy Σ Γ) :=
  let Q := Build_Alphabet (Mealy_st M) Mealy_init in
  @tm_fix Σ (Al_prod Γ Q) (Al_pt Γ, Mealy_init)
       (@tm_memoryless (Al_prod Σ (Al_prod Γ Q)) (Al_prod Γ Q) (fun z => let '(a,(_,q)) := z in Mealy_delta (a,q))).

Definition tm_of_Mealy {Σ Γ} (M : Mealy Σ Γ) :=
  let Q := Build_Alphabet (Mealy_st M) Mealy_init in
  tm_comp (@tm_memoryless (Al_prod Γ Q) Γ fst)
          (tm_run_of_Mealy M).

Lemma tm_run_of_Mealy_eq : forall (Σ Γ : Alphabet) (M : Mealy Σ Γ),
  let Q := Build_Alphabet (Mealy_st M) Mealy_init in
  forall u, run_of_Mealy M u == fun_of_tm (tm_run_of_Mealy M) u.
  intros.
  unfold tm_run_of_Mealy, fun_of_tm.
  simpl.
  refine (@Mealy_fix_fun_uniq Σ (Al_prod Γ Q) (Mealy_memoryless _) (Al_pt Γ, Mealy_init) u (run_of_Mealy M u) _ x).
  clear x; intros.
  rewrite Mealy_memoryless_fun_eq.
  destruct x; reflexivity.
Qed.

(* not an iso; the state space gets multiplied by Γ *)
Lemma Mealy_of_tm_of_Mealy_morph : forall (Σ Γ : Alphabet) (M : Mealy Σ Γ),
  Mealy_morph (Mealy_of_tm (tm_of_Mealy M)) M.
  intros; unshelve esplit.
  - simpl.
    exact (fun z => let '(tt, (_, q), tt) := z in q).
  - reflexivity.
  - intros; simpl in *. destruct q as [[[] [? ?]] []].
    destruct (Mealy_delta (a, f0)); reflexivity.
Qed.

Lemma tm_Mealy_fun_eq : forall (Σ Γ : Alphabet) (M : Mealy Σ Γ),
  forall u, fun_of_tm (tm_of_Mealy M) u == fun_of_Mealy M u.
  intros; unfold fun_of_tm.
  apply Mealy_morph_fun_eq, Mealy_of_tm_of_Mealy_morph.
Qed.


(* Now that we have checked expressivity of the term language,
we should start axiomatizing its equational theory *)

Reserved Notation "t ≡ u" (at level 54).
Notation "t ∘ u" := (@tm_comp _ _ _ t u) (at level 23, left associativity).
Notation "[ f ]ω" := (@tm_memoryless _ _ f).

Lemma tm_memoryless_fun_eq : forall (Σ Γ : Alphabet) (f : Σ -> Γ),
  forall w, fun_of_tm [f]ω w == (fun n => f (w n)).
  intros; unfold fun_of_tm; simpl.
  apply Mealy_memoryless_fun_eq.
Qed.

Lemma tm_comp_fun_eq : forall (Σ Γ Θ : Alphabet) (t : tm Γ Θ) (u : tm Σ Γ),
  forall w, fun_of_tm (t ∘ u) w == (fun_of_tm t) (fun_of_tm u w).
  intros; unfold fun_of_tm; simpl.
  apply Mealy_comp_fun_eq.
Qed.


Definition tm_cons {Σ : Alphabet} (a₀ : Σ) : tm Σ Σ :=
 @tm_comp Σ Σ (Al_prod Σ Σ) (@tm_memoryless (Al_prod Σ Σ) Σ snd)
    (@tm_fix Σ (Al_prod Σ Σ) (a₀, a₀) [fun (z : Al_prod _ (Al_prod _ _)) =>
       let '(a, (b, _)) := z in (a,b) : Al_prod _ _]ω).


Fixpoint tm_times_id {Σ Γ : Alphabet} (t : tm Σ Γ) Θ : tm (Al_prod Σ Θ) (Al_prod Γ Θ) :=
  match t with
   [f]ω => [fun (a : Σ × Θ) => let (x,y) return Γ × Θ := a in (f x, y)]ω
 | t ∘ t' => (tm_times_id t Θ) ∘ (tm_times_id t' Θ)
 | tm_fix b₀ t => tm_fix ((b₀, Al_pt Θ) : Γ × Θ)
         ((tm_times_id t Θ) ∘
             [fun z : (Σ × Θ) × (Γ × Θ) =>
                  let '((a,c),(b,_)) := z in ((a,b),c) : (Σ × Γ) × Θ]ω)
  end.

Definition tm_twist {Σ Γ : Alphabet} : tm (Σ × Γ) (Γ × Σ) :=
  [fun z : Σ × Γ => let '(x, y) := z in (y,x) : Γ × Σ ]ω.

Definition tm_delta {Σ : Alphabet} : tm Σ (Σ × Σ) :=
  [fun z : Σ => (z,z) : Σ × Σ ]ω.

Definition id_times_tm {Σ Γ : Alphabet} Θ (t : tm Σ Γ) : tm (Θ × Σ) (Θ × Γ) :=
  tm_twist ∘ (tm_times_id t Θ) ∘ tm_twist.

Definition tm_times' {Σ Γ Θ Δ : Alphabet} (t : tm Σ Θ) (u : tm Γ Δ) : tm (Σ × Γ) (Θ × Δ) :=
  (id_times_tm _ u) ∘ (tm_times_id t _).

Definition tm_pair {Σ Γ Θ : Alphabet} (t : tm Σ Γ) (u : tm Σ Θ) : tm Σ (Γ × Θ) :=
  tm_times' t u ∘ tm_delta.

Definition tm_proj1 {Σ Γ : Alphabet} : tm (Σ × Γ) Σ := [fun x : Σ × Γ => fst x]ω.

Definition tm_proj2 {Σ Γ : Alphabet} : tm (Σ × Γ) Γ :=
  [fun x : Σ × Γ => snd x]ω.

Definition tm_times {Σ Γ Θ Δ : Alphabet} (t : tm Σ Θ) (u : tm Γ Δ) : tm (Σ × Γ) (Θ × Δ) :=
  tm_pair (t ∘ tm_proj1) (u ∘ tm_proj2).

Definition tm_id {Σ : Alphabet} : tm Σ Σ := [fun x => x]ω.

Lemma tm_fix_fun_eq : forall (Σ Γ : Alphabet) (b₀ : Γ) (t : tm (Σ × Γ) Γ),
  forall w, fun_of_tm (tm_fix b₀ t) w == fun_of_tm t (pairfun w (str_cons b₀ (fun_of_tm (tm_fix b₀ t) w))).
  intros; unfold fun_of_tm; simpl.
  apply Mealy_fix_fun_eq.
Qed.

Lemma tm_fix_fun_uniq : forall (Σ Γ : Alphabet) (b₀ : Γ) (t : tm (Σ × Γ) Γ),
  forall u w, u == fun_of_tm t (pairfun w (str_cons b₀ u)) ->
   u == fun_of_tm (tm_fix b₀ t) w.
  unfold fun_of_tm; simpl; intros; apply Mealy_fix_fun_uniq; assumption.
Qed.

Lemma fun_of_tm_ext : forall (Σ Γ : Alphabet) (t : tm Σ Γ),
  forall u u', u == u' -> fun_of_tm t u == fun_of_tm t u'.
  intros; unfold fun_of_tm; erewrite fun_of_Mealy_ext; eauto.
Qed.

Fixpoint fun0_of_tm {Σ Γ : Alphabet} (t : tm Σ Γ) (a : Σ) : Γ :=
  match t with
   [f]ω => f a
 | t ∘ u => fun0_of_tm t (fun0_of_tm u a)
 | tm_fix b₀ t => fun0_of_tm t (a,b₀)
  end.

Fixpoint next_tm {Σ Γ : Alphabet} (t : tm Σ Γ) (a₀ : Σ) : tm Σ Γ :=
  match t with
   [f]ω => [f]ω
 | t ∘ u => next_tm t (fun0_of_tm u a₀) ∘ next_tm u a₀
 | tm_fix b₀ t => tm_fix (fun0_of_tm t (a₀,b₀)) (next_tm t (a₀,b₀))
  end.
