Require Import Bool.
Require Import List.
Require Import finset.
Require Import mealy.
Require Import tm.

(* The equational theory of terms presented as a derivation tree.
Nodes may be created to ensure that:
- ≡ is an equivalence relation
- ≡ is a congruence relation for the syntax
- [f]ω ≡ [g]ω whenever f == g
- [-]ω is functorial
- ∘ is associative, with units given by tm_id
- tm_fix b₀ t is a fixpoint of *
- all fixpoints of * are tm_fix b₀ t
- cartesian structure given by tm_pair, tm_proj1, tm_proj2
- ≡ is a congruence for pairing
- tm_cons a₀ is natural against memoryless functions, i.e
  [f]ω ∘ cons a₀ ≡ cons (f a₀) ∘ [f]ω

* is the following equation
X ≡ t ∘ tm_pair tm_id (tm_cons b₀ ∘ X)
*)

Inductive tm_eq : forall Σ Γ : Alphabet, tm Σ Γ -> tm Σ Γ -> Type :=
  tm_eq_refl : forall Σ Γ (t : tm Σ Γ), t ≡ t
| tm_eq_sym : forall Σ Γ (t u : tm Σ Γ), t ≡ u -> u ≡ t
| tm_eq_trans : forall Σ Γ (t u v : tm Σ Γ), t ≡ u -> u ≡ v -> t ≡ v
| tm_eq_comp_cong : forall Σ Γ {Θ : Alphabet} (t t' : tm Θ Γ) (u u' : tm Σ Θ),
   t ≡ t' -> u ≡ u' -> t ∘ u ≡ t' ∘ u'
| tm_eq_fix_cong : forall Σ Γ (t t' : tm (Σ × Γ) Γ) (a₀ : Γ), t ≡ t' -> tm_fix a₀ t ≡ tm_fix a₀ t'
| tm_eq_memoryless : forall (Σ Γ : Alphabet) (f g : Σ -> Γ), f == g -> [f]ω ≡ [g]ω
| tm_eq_memoryless_comp : forall (Σ Γ Θ : Alphabet) (f : Θ -> Γ) (g : Σ -> Θ),
                          [fun x : Σ => f (g x)]ω ≡ ([f]ω ∘ [g]ω)
| tm_eq_comp_assoc : forall (Σ Γ Θ Θ' : Alphabet) (t : tm Θ' Γ) (u : tm Θ Θ') (v : tm Σ Θ),
    ((t ∘ u) ∘ v) ≡ (t ∘ (u ∘ v))
| tm_eq_comp_unitl : forall Σ Γ (t : tm Σ Γ), tm_id ∘ t ≡ t
| tm_eq_comp_unitr : forall Σ Γ (t : tm Σ Γ), t ∘ tm_id ≡ t
| tm_eq_fix_unfold : forall Σ Γ b₀ (t : tm (Al_prod Σ Γ) Γ),
   tm_fix b₀ t ≡ t ∘ tm_pair tm_id (tm_cons b₀ ∘ (tm_fix b₀ t))
| tm_eq_fix_uniq : forall Σ Γ b₀ (t : tm (Al_prod Σ Γ) Γ) (u : tm Σ Γ),
   u ≡ t ∘ tm_pair tm_id (tm_cons b₀ ∘ u) -> u ≡ tm_fix b₀ t
| tm_eq_proj1_pair : forall Σ Γ Θ (t : tm Σ Γ) (u : tm Σ Θ), tm_proj1 ∘ tm_pair t u ≡ t
| tm_eq_proj2_pair : forall Σ Γ Θ (t : tm Σ Γ) (u : tm Σ Θ), tm_proj2 ∘ tm_pair t u ≡ u
| tm_eq_surj_pair : forall Σ Γ, @tm_pair (Σ × Γ) Σ Γ tm_proj1 tm_proj2 ≡ tm_id
| tm_eq_pair_compr : forall Σ Σ' Γ Θ (t : tm Σ Γ) (u : tm Σ Θ) (v : tm Σ' Σ), tm_pair (t ∘ v) (u ∘ v) ≡ tm_pair t u ∘ v
| tm_eq_pair_cong : forall Σ Γ Θ (t t' : tm Σ Γ) (u u' : tm Σ Θ), t ≡ t' -> u ≡ u' -> tm_pair t u ≡ tm_pair t' u'
| tm_eq_cons_nat_memoryless : forall (Σ Γ : Alphabet) (f : Σ -> Γ) a₀, [f]ω ∘ tm_cons a₀ ≡ tm_cons (f a₀) ∘ [f]ω
where "t ≡ u" := (@tm_eq _ _ t u).

Arguments tm_eq_comp_cong [_ _ _ _ _ _ _] _ _.
Arguments tm_eq_pair_cong [_ _ _ _ _ _ _] _ _.
Arguments tm_eq_fix_cong [_ _ _ _ _] _.
Arguments tm_eq_refl [_ _ _].
Arguments tm_eq_sym [_ _ _ _] _.
Arguments tm_eq_trans [_ _ _ _ _] _ _.
Arguments tm_eq_comp_assoc [_ _ _ _ _ _ _].
Arguments tm_eq_comp_unitl [_ _ _].
Arguments tm_eq_comp_unitr [_ _ _].
Arguments tm_eq_fix_unfold [_ _ _ _].
Arguments tm_eq_proj1_pair [_ _ _ _ _].
Arguments tm_eq_proj2_pair [_ _ _ _ _].
Arguments tm_eq_surj_pair [_ _].
Arguments tm_eq_pair_cong [_ _ _ _ _ _ _].
Arguments tm_eq_pair_compr [_ _ _ _ _ _ _].
Arguments tm_eq_cons_nat_memoryless [_ _ _ _].
Arguments tm_eq_memoryless_comp [_ _ _ _ _].

Definition tm_eq_comp_congl : forall {Σ Γ Θ : Alphabet} {t t' : tm Θ Γ} {u : tm Σ Θ},
  t ≡ t' -> t ∘ u ≡ t' ∘ u := fun _ _ _ _ _ _ p => tm_eq_comp_cong p tm_eq_refl.
Definition tm_eq_comp_congr : forall {Σ Γ Θ : Alphabet} {t : tm Θ Γ} {u u' : tm Σ Θ},
  u ≡ u' -> t ∘ u ≡ t ∘ u' := fun _ _ _ _ _ _ p => tm_eq_comp_cong tm_eq_refl p.
Definition tm_eq_pair_congl : forall {Σ Γ Θ : Alphabet} {t t' : tm Σ Γ} {u : tm Σ Θ},
  t ≡ t' -> tm_pair t u ≡ tm_pair t' u := fun _ _ _ _ _ _ p => tm_eq_pair_cong p tm_eq_refl.
Definition tm_eq_pair_congr : forall {Σ Γ Θ : Alphabet} {t : tm Σ Γ} {u u' : tm Σ Θ},
  u ≡ u' -> tm_pair t u ≡ tm_pair t u' := fun _ _ _ _ _ _ p => tm_eq_pair_cong tm_eq_refl p.

Hint Constructors tm_eq : mealy.
Remove Hints tm_eq_sym : mealy.
Hint Rewrite fun_of_tm_ext : mealy.
Hint Rewrite tm_memoryless_fun_eq : mealy.
Hint Rewrite tm_comp_fun_eq : mealy.

Lemma tm_times_id_fun_eq : forall (Σ Γ Θ : Alphabet) (t : tm Σ Γ),
  forall w, fun_of_tm (tm_times_id t Θ) w == (fun n => (fun_of_tm t (fun n => fst (w n)) n, snd (w n))).
   induction t; intros; simpl.
   - repeat rewrite tm_memoryless_fun_eq.
     destruct (w x); reflexivity.
   - do 2 rewrite tm_comp_fun_eq. rewrite IHt1, IHt2.
     simpl.
     destruct (w x).
     apply f_equal with (f := (fun x => (x,f0))).
     apply fun_of_tm_ext.
     intro n; simpl.
     rewrite IHt2; reflexivity.
   - symmetry.
     apply (tm_fix_fun_uniq (Σ × Θ) (Γ × Θ) (b₀, Al_pt Θ) ((tm_times_id t Θ) ∘
             [fun z : (Σ × Θ) × (Γ × Θ) =>
                  let '((a,c),(b,_)) := z in ((a,b),c) : (Σ × Γ) × Θ]ω)
            (fun x => (fun_of_tm (tm_fix b₀ t) (fun n : nat => fst (w n)) x, snd (w x)) )).
     clear x; intros n.
     rewrite tm_comp_fun_eq, IHt; simpl.
     apply f_equal2.
     + rewrite tm_fix_fun_eq.
       apply fun_of_tm_ext; clear n; intros n.
       unfold pairfun; simpl.
       rewrite tm_memoryless_fun_eq; simpl.
       destruct (w n); simpl.
       destruct n; reflexivity.
     + rewrite tm_memoryless_fun_eq; simpl.
       destruct (w n); simpl.
       destruct n; reflexivity.
Qed.

Lemma id_times_tm_fun_eq : forall (Σ Γ Θ : Alphabet) (t : tm Σ Γ),
  forall w, fun_of_tm (id_times_tm Θ t) w == (fun n => (fst (w n), fun_of_tm t (fun n => snd (w n)) n)).
   intros; unfold id_times_tm, tm_twist.
   repeat rewrite tm_comp_fun_eq || rewrite tm_memoryless_fun_eq.
   rewrite tm_times_id_fun_eq; simpl.
   repeat rewrite tm_comp_fun_eq || rewrite tm_memoryless_fun_eq.
   apply f_equal2; [destruct (w x); reflexivity|].
   apply fun_of_tm_ext; intros n; rewrite tm_memoryless_fun_eq.
   destruct (w n); reflexivity.
Qed.

Definition timesfun {Σ Γ Σ' Γ'} (f : Σ -> Γ) (g : Σ' -> Γ') : Σ * Σ' -> Γ * Γ' :=
  fun '(a,b) => (f a, g b).

Definition deltafun {Σ} (a : Σ) := (a,a).

Definition twistfun {Σ Γ} : Σ * Γ -> Γ * Σ :=
  fun '(a,b) => (b,a).

Lemma pairfun_decomp_eq {Σ Γ Θ} : forall (f : Σ -> Γ) (g : Σ -> Θ), 
  forall a, pairfun f g a = timesfun id g (timesfun f id (deltafun a)).
  intros; reflexivity.
Qed.

Lemma tm_pair_fun_eq : forall (Σ Γ Θ : Alphabet) (t : tm Σ Γ) (u : tm Σ Θ) w,
  fun_of_tm (tm_pair t u) w == pairfun (fun_of_tm t w) (fun_of_tm u w).
   intros.
   unfold tm_pair, tm_times'.
   repeat rewrite tm_comp_fun_eq.
   rewrite pairfun_decomp_eq.
   rewrite id_times_tm_fun_eq; simpl.
   rewrite tm_times_id_fun_eq; simpl.
   unfold id. apply f_equal2; apply fun_of_tm_ext; intros n.
   - unfold tm_delta; rewrite tm_memoryless_fun_eq; reflexivity.
   - rewrite tm_times_id_fun_eq; simpl.
     unfold tm_delta; rewrite tm_memoryless_fun_eq; reflexivity.
Qed.

Lemma tm_id_fun_eq : forall Σ w, fun_of_tm (@tm_id Σ) w == w.
  intros; unfold tm_id; rewrite tm_memoryless_fun_eq; reflexivity.
Qed.

Lemma tm_proj1_fun_eq : forall Σ Γ, forall w, fun_of_tm (@tm_proj1 Σ Γ) w == (fun n => fst (w n)).
  intros; unfold tm_proj1; rewrite tm_memoryless_fun_eq; reflexivity.
Qed.

Lemma tm_proj2_fun_eq : forall Σ Γ, forall w, fun_of_tm (@tm_proj2 Σ Γ) w == (fun n => snd (w n)).
  intros; unfold tm_proj2; rewrite tm_memoryless_fun_eq; reflexivity.
Qed.

Lemma tm_cons_fun_eq : forall {Σ : Alphabet} (a₀ : Σ) w, fun_of_tm (tm_cons a₀) w == str_cons a₀ w.
  unfold tm_cons, fun_of_tm; simpl.
  intros; rewrite <-Mealy_cons_fun_eq.
  reflexivity.
Qed.

Lemma tm_fix_ext : forall {Σ Γ : Alphabet} {b₀ : Γ} {t u : tm (Σ × Γ) _},
  (forall w, fun_of_tm t w == fun_of_tm u w) -> forall w, fun_of_tm (tm_fix b₀ t) w == fun_of_tm (tm_fix b₀ u) w.
  intros; apply tm_fix_fun_uniq.
  intros; symmetry.
  rewrite <-H; symmetry; apply tm_fix_fun_eq.
Qed.

Theorem soundness : forall (Σ Γ : Alphabet) (t u : tm Σ Γ), t ≡ u -> forall w, fun_of_tm t w == fun_of_tm u w.
  intros ? ? ? ? p; induction p; intros; auto.
  - rewrite IHp1; apply IHp2.
  - repeat rewrite tm_comp_fun_eq; rewrite IHp1.
    apply fun_of_tm_ext; apply IHp2.
  - apply tm_fix_ext; auto.
  - do 2 rewrite tm_memoryless_fun_eq; apply e.
  - repeat rewrite tm_comp_fun_eq || rewrite tm_memoryless_fun_eq; reflexivity.
  - repeat rewrite tm_comp_fun_eq.
    apply fun_of_tm_ext.
    intro; rewrite tm_comp_fun_eq; reflexivity.
  - rewrite tm_comp_fun_eq.
    rewrite tm_id_fun_eq; reflexivity.
  - rewrite tm_comp_fun_eq.
    apply fun_of_tm_ext.
    intro; rewrite tm_id_fun_eq; reflexivity.
  - rewrite tm_fix_fun_eq, tm_comp_fun_eq.
    apply fun_of_tm_ext.
    intros n; clear x; rewrite tm_pair_fun_eq.
    unfold pairfun; simpl; rewrite tm_id_fun_eq; apply f_equal.
    rewrite tm_comp_fun_eq.
    rewrite tm_cons_fun_eq; reflexivity.
  - apply tm_fix_fun_uniq.
    clear x; intro n.
    rewrite (IHp w n).
    rewrite tm_comp_fun_eq.
    apply fun_of_tm_ext.
    intros k.
    rewrite tm_pair_fun_eq; unfold pairfun; simpl.
    rewrite tm_id_fun_eq; apply f_equal.
    rewrite tm_comp_fun_eq, tm_cons_fun_eq.
    destruct k; auto.
  - rewrite tm_comp_fun_eq, tm_proj1_fun_eq, tm_pair_fun_eq; reflexivity.
  - rewrite tm_comp_fun_eq, tm_proj2_fun_eq, tm_pair_fun_eq; reflexivity.
  - rewrite tm_pair_fun_eq; unfold pairfun.
    rewrite tm_proj1_fun_eq, tm_proj2_fun_eq, tm_id_fun_eq.
    destruct (w x); reflexivity.
  - rewrite tm_pair_fun_eq; unfold pairfun; repeat rewrite tm_comp_fun_eq.
    rewrite tm_pair_fun_eq; reflexivity.  
  - repeat rewrite tm_pair_fun_eq; unfold pairfun; simpl.
    rewrite IHp1, IHp2; reflexivity.
  - repeat rewrite tm_comp_fun_eq;
    rewrite tm_cons_fun_eq,
            tm_memoryless_fun_eq,
            tm_cons_fun_eq;unfold str_cons;simpl.
    destruct x; try rewrite tm_memoryless_fun_eq;reflexivity.
Qed.

Fixpoint memoryless {Σ Γ : Alphabet} (t : tm Σ Γ) : bool :=
  match t with
   tm_fix _ _ => false
 | t ∘ u =>  memoryless t && memoryless u
 | _ => true
  end.

Lemma memoryless_eq_normal : forall {Σ Γ : Alphabet} (t : tm Σ Γ),
 memoryless t -> t ≡ [fun0_of_tm t]ω.
  intros; induction t; simpl in *.
  - apply tm_eq_memoryless; auto.
  - apply andb_prop in H.
    destruct H as [H H']; specialize (IHt1 H); specialize (IHt2 H').
    eapply tm_eq_trans.
    + apply tm_eq_comp_cong; eassumption.
    + apply tm_eq_sym, tm_eq_memoryless_comp.
  - inversion H.
Qed.

Lemma fun_fun0_of_tm_eq : forall (Σ Γ : Alphabet) (t : tm Σ Γ) u,
  fun_of_tm t u 0 = fun0_of_tm t (u 0).
  induction t; intros.
  - rewrite tm_memoryless_fun_eq; reflexivity.
  - rewrite tm_comp_fun_eq, IHt1, IHt2;reflexivity.
  - rewrite tm_fix_fun_eq, IHt; reflexivity.
Qed.

Lemma memoryless_eq_complete : forall (Σ Γ : Alphabet) (t u : tm Σ Γ),
  memoryless t -> memoryless u -> (forall w, fun_of_tm t w == fun_of_tm u w) -> t ≡ u.
  intros.
  apply (tm_eq_trans (memoryless_eq_normal t H)), tm_eq_sym.
  apply (tm_eq_trans (memoryless_eq_normal u H0)).
  apply tm_eq_memoryless; intros a.
  assert(P := H1 (fun _ => a) 0).
  do 2 rewrite fun_fun0_of_tm_eq in P; auto.
Qed.

Ltac solve_tm_eq_memoryless :=
  try (apply memoryless_eq_complete; try reflexivity); intros;
  unfold tm_id, tm_proj1, tm_proj2, tm_pair, tm_times, tm_times', tm_times_id, id_times_tm, tm_twist, tm_delta; simpl;
  repeat (rewrite tm_memoryless_fun_eq ||
          rewrite tm_comp_fun_eq); auto.

Lemma tm_times_in_tm_pair : forall {Σ Σ' Γ Γ'} {t : tm Σ Σ'} {t' : tm Γ Γ'},
  tm_times t t' ≡ tm_pair (t ∘ tm_proj1) (t' ∘ tm_proj2).
  intros; apply tm_eq_refl.
Qed.

Lemma tm_times_tm_pair : forall {Σ Σ' Γ Γ' Θ} {u : tm Θ Σ} {v : tm Θ Γ} {t : tm Σ Σ'} {t' : tm Γ Γ'}, tm_times t t' ∘ tm_pair u v ≡ tm_pair (t ∘ u) (t' ∘ v).
  intros.
  unfold tm_times.
  apply (tm_eq_trans (tm_eq_sym tm_eq_pair_compr)).
  apply tm_eq_pair_cong; apply (tm_eq_trans ( tm_eq_comp_assoc)), tm_eq_comp_congr;
  [apply tm_eq_proj1_pair|apply tm_eq_proj2_pair].
Qed.

Lemma tm_fix_pre_comp : forall {Σ Γ Θ : Alphabet} {t : tm (Θ × Γ) Γ} {u : tm Σ Θ} {b₀},
  (tm_fix b₀ t) ∘ u ≡ tm_fix b₀ (t ∘ tm_times u tm_id).
  intros.
  apply tm_eq_fix_uniq.
  eapply tm_eq_trans; [apply tm_eq_comp_cong; [apply tm_eq_fix_unfold | apply tm_eq_refl] |].
  apply (tm_eq_trans tm_eq_comp_assoc).
  apply tm_eq_sym.
  apply (tm_eq_trans tm_eq_comp_assoc).
  apply tm_eq_comp_cong; [apply tm_eq_refl|].
  apply (tm_eq_trans tm_times_tm_pair), tm_eq_sym.
  apply (tm_eq_trans (tm_eq_sym tm_eq_pair_compr)).
  apply tm_eq_pair_cong.
  + apply (tm_eq_trans (tm_eq_sym tm_eq_comp_unitr)),
          (tm_eq_trans tm_eq_comp_assoc),
          tm_eq_comp_unitl.
  + apply tm_eq_sym,
          (tm_eq_trans tm_eq_comp_unitl),
          (tm_eq_sym tm_eq_comp_assoc).
Qed.

Lemma tm_fix_post_comp_memoryless : forall {Σ Γ Θ : Alphabet} {t : tm (Σ × Θ) Γ} {f : Γ -> Θ} {b₀}, [f]ω ∘ tm_fix b₀ (t ∘ tm_times tm_id [f]ω) ≡ tm_fix (f b₀) ([f]ω ∘ t).
  intros; apply tm_eq_fix_uniq.
  apply tm_eq_sym, (tm_eq_trans tm_eq_comp_assoc), tm_eq_comp_congr.
  apply tm_eq_sym, (tm_eq_trans tm_eq_fix_unfold).
  apply (tm_eq_trans tm_eq_comp_assoc).
  apply tm_eq_comp_congr, (tm_eq_trans tm_times_tm_pair).
  apply tm_eq_pair_cong; [apply tm_eq_comp_unitl|].
  apply (tm_eq_trans (tm_eq_sym tm_eq_comp_assoc)),
        tm_eq_sym,
        (tm_eq_trans (tm_eq_sym tm_eq_comp_assoc)).
  apply tm_eq_comp_congl, tm_eq_sym, tm_eq_cons_nat_memoryless.
Qed.

Definition tm_assoc {Σ Γ Θ : Alphabet} : tm ((Σ × Γ) × Θ) (Σ × (Γ × Θ)) :=
  tm_pair (tm_proj1 ∘ tm_proj1) (tm_pair (tm_proj2 ∘ tm_proj1) tm_proj2).

Definition tm_assoc_inv {Σ Γ Θ : Alphabet} : tm (Σ × (Γ × Θ)) ((Σ × Γ) × Θ) :=
  tm_pair (tm_pair tm_proj1 (tm_proj1 ∘ tm_proj2)) (tm_proj2 ∘ tm_proj2).

Lemma tm_fix_fix : forall {Σ Γ : Alphabet} {t : tm ((Σ × Γ) × Γ) Γ} {q₀ q₁},
 tm_delta ∘ tm_fix q₀ (tm_fix q₁ t) ≡ tm_fix ((q₀,q₁) : Γ × Γ) (tm_delta ∘ t ∘ tm_assoc_inv).
  intros; apply tm_eq_fix_uniq.
  apply tm_eq_sym.
  repeat apply (tm_eq_trans tm_eq_comp_assoc).
  apply tm_eq_comp_congr.
  apply tm_eq_sym, (tm_eq_trans tm_eq_fix_unfold).
  apply (tm_eq_trans (tm_eq_comp_congl tm_eq_fix_unfold)).
  repeat apply (tm_eq_trans tm_eq_comp_assoc); apply tm_eq_comp_congr.
  apply (tm_eq_trans (tm_eq_sym tm_eq_pair_compr)).
  unfold tm_assoc_inv; apply tm_eq_sym.
  apply (tm_eq_trans (tm_eq_sym tm_eq_pair_compr)).
  apply tm_eq_pair_cong.
  + apply (tm_eq_trans (tm_eq_sym tm_eq_pair_compr)).
    apply tm_eq_sym, (tm_eq_trans tm_eq_comp_unitl).
    apply tm_eq_pair_cong.
    - apply tm_eq_sym, tm_eq_proj1_pair.
    - apply tm_eq_sym, (tm_eq_trans tm_eq_comp_assoc).
      apply (tm_eq_trans (tm_eq_comp_congr tm_eq_proj2_pair)).
      apply (tm_eq_trans (tm_eq_sym tm_eq_comp_assoc)),
            (tm_eq_trans (tm_eq_comp_congl tm_eq_cons_nat_memoryless)).
      apply (tm_eq_trans (tm_eq_comp_assoc)).
      apply tm_eq_comp_congr.
      apply (tm_eq_trans (tm_eq_sym tm_eq_comp_assoc)),
            tm_eq_sym, (tm_eq_trans (tm_eq_sym tm_eq_comp_unitl)),
            tm_eq_comp_congl.
      solve_tm_eq_memoryless.
  + apply (tm_eq_trans tm_eq_comp_assoc).
    apply (tm_eq_trans (tm_eq_comp_congr tm_eq_proj2_pair)).
    apply (tm_eq_trans (tm_eq_sym tm_eq_comp_assoc)),
          (tm_eq_trans (tm_eq_comp_congl tm_eq_cons_nat_memoryless));simpl.
    apply (tm_eq_trans tm_eq_comp_assoc),
          tm_eq_sym,
          (tm_eq_trans tm_eq_comp_assoc),
          tm_eq_comp_congr. 
    apply (tm_eq_trans (tm_eq_sym tm_eq_fix_unfold)),
          (tm_eq_trans (tm_eq_sym tm_eq_comp_unitl)).
    apply tm_eq_sym, (tm_eq_trans (tm_eq_sym tm_eq_comp_assoc)), tm_eq_comp_congl.
    solve_tm_eq_memoryless.
Qed.    

Corollary tm_fix_fix_proj1 : forall {Σ Γ : Alphabet} {t : tm ((Σ × Γ) × Γ) Γ} {q₀ q₁},
 tm_fix q₀ (tm_fix q₁ t) ≡ tm_proj1 ∘ tm_fix ((q₀,q₁) : Γ × Γ) (tm_delta ∘ t ∘ tm_assoc_inv).
  intros.
  apply (tm_eq_trans (tm_eq_sym tm_eq_comp_unitl)).
  refine (@tm_eq_trans _ _ _ ((tm_proj1 ∘ tm_delta) ∘ tm_fix q₀ (tm_fix q₁ t)) _ _ _).
  + apply tm_eq_comp_congl.
    solve_tm_eq_memoryless.
  + apply (tm_eq_trans tm_eq_comp_assoc), tm_eq_comp_congr.
    apply tm_fix_fix.
Qed.

Lemma tm_fix_comp_fix : forall {Σ Γ Θ : Alphabet} {t : tm (Σ × Γ) Γ} {u : tm (Γ × Θ) Θ} {a₀} {b₀}, tm_pair tm_id (tm_fix b₀ u) ∘ tm_fix a₀ t ≡ tm_fix ((a₀, b₀) : _ × _) (tm_pair tm_proj1 u ∘ tm_times t tm_id ∘ tm_assoc_inv).
  intros.
  assert (t ∘ tm_pair tm_proj1 (tm_proj1 ∘ tm_proj2) ∘ tm_pair tm_id (tm_cons ((a₀, b₀) : _ × _) ∘ (tm_pair tm_id (tm_fix b₀ u) ∘ tm_fix a₀ t)) ≡ tm_id ∘ tm_fix a₀ t).
  {
       apply tm_eq_sym, (tm_eq_trans tm_eq_comp_unitl).
       apply (tm_eq_trans tm_eq_fix_unfold).
       apply tm_eq_sym, (tm_eq_trans tm_eq_comp_assoc).
       apply tm_eq_comp_congr.       
       apply (tm_eq_trans (tm_eq_sym tm_eq_pair_compr)).
       apply tm_eq_pair_cong.
       * apply tm_eq_proj1_pair.
       * apply (tm_eq_trans tm_eq_comp_assoc),
               (tm_eq_trans (tm_eq_comp_congr tm_eq_proj2_pair)),
               (tm_eq_trans (tm_eq_sym tm_eq_comp_assoc)).
         apply (tm_eq_trans (tm_eq_comp_congl tm_eq_cons_nat_memoryless)),
               (tm_eq_trans tm_eq_comp_assoc), tm_eq_comp_congr.
         apply (tm_eq_trans (tm_eq_comp_congr (tm_eq_sym tm_eq_pair_compr))),
               (tm_eq_trans tm_eq_proj1_pair), tm_eq_comp_unitl.
  }
  apply tm_eq_fix_uniq.
  apply (tm_eq_trans (tm_eq_sym tm_eq_pair_compr)).
  apply tm_eq_sym.
  do 2 apply (tm_eq_trans tm_eq_comp_assoc).
  apply (tm_eq_trans (tm_eq_sym tm_eq_pair_compr)).
  apply tm_eq_pair_cong.
  + do 2 apply (tm_eq_trans (tm_eq_sym tm_eq_comp_assoc)).
    refine (tm_eq_trans (tm_eq_comp_congl _) _).
    - apply (tm_eq_trans tm_eq_comp_assoc).
      apply (tm_eq_trans (tm_eq_comp_congr tm_times_tm_pair)).
      apply (tm_eq_trans tm_eq_proj1_pair), tm_eq_refl.
    - apply X.
  + apply tm_eq_sym, (tm_eq_trans (tm_eq_comp_congl tm_eq_fix_unfold)).
    apply (tm_eq_trans tm_eq_comp_assoc).
    apply tm_eq_comp_congr.
    apply (tm_eq_trans (tm_eq_sym tm_eq_pair_compr)).
    unfold tm_assoc_inv.    
    apply tm_eq_sym, (tm_eq_trans (tm_eq_sym tm_eq_comp_assoc)).
    apply (tm_eq_trans (tm_eq_comp_congl tm_times_tm_pair)),
          (tm_eq_trans (tm_eq_sym tm_eq_pair_compr)).
    apply tm_eq_pair_cong.
    - apply X.
    - apply (tm_eq_trans (tm_eq_comp_congl tm_eq_comp_unitl)).
      apply (tm_eq_trans tm_eq_comp_assoc),
            (tm_eq_trans (tm_eq_comp_congr tm_eq_proj2_pair)).
      apply (tm_eq_trans (tm_eq_sym tm_eq_comp_assoc)),
            (tm_eq_trans (tm_eq_comp_congl tm_eq_cons_nat_memoryless)),
            (tm_eq_trans (tm_eq_sym tm_eq_comp_assoc)),
            tm_eq_comp_congl.
      apply (tm_eq_trans tm_eq_comp_assoc), tm_eq_comp_congr.
      apply tm_eq_proj2_pair.
Qed.

Corollary tm_fix_comp_fix_proj2 : forall {Σ Γ Θ : Alphabet} {t : tm (Σ × Γ) Γ} {u : tm (Γ × Θ) Θ} {a₀} {b₀}, tm_fix b₀ u ∘ tm_fix a₀ t ≡ tm_proj2 ∘ tm_fix ((a₀, b₀) : _ × _) (tm_pair tm_proj1 u ∘ tm_times t tm_id ∘ tm_assoc_inv).
  intros.
  apply tm_eq_sym.  
  apply (tm_eq_trans (tm_eq_comp_congr (tm_eq_sym tm_fix_comp_fix))).
  apply (tm_eq_trans (tm_eq_sym tm_eq_comp_assoc)),
        (tm_eq_comp_congl tm_eq_proj2_pair).
Qed.

Lemma tm_fix_const : forall {Σ Γ : Alphabet} {t : tm Σ Γ} (b₀ : Γ), t ≡ tm_fix b₀ (t ∘ tm_proj1).
  intros; apply tm_eq_fix_uniq.
  apply (tm_eq_trans (tm_eq_sym tm_eq_comp_unitr)),
        tm_eq_sym,
        (tm_eq_trans tm_eq_comp_assoc),
        tm_eq_comp_congr, tm_eq_proj1_pair.
Qed.

Lemma tm_fix_memoryless : forall {Σ Γ : Alphabet} {f} (b₀ : Γ), [f]ω ≡ tm_fix b₀ [fun (z : Σ × Γ) => f (fst z)]ω.
  intros; apply (tm_eq_trans (tm_fix_const b₀)), tm_eq_fix_cong.
  solve_tm_eq_memoryless.
Qed.

Lemma tm_fix_inner_memoryless_comp : forall {Σ Γ Θ : Alphabet} {t : tm (Σ × Θ) Γ} {f : Γ -> Θ} {c₀}, tm_fix (f c₀) ([f]ω ∘ t) ≡ [f]ω ∘ tm_fix c₀ (t ∘ tm_times tm_id [f]ω).
  intros; apply tm_eq_sym, tm_eq_fix_uniq.
  apply tm_eq_sym, (tm_eq_trans tm_eq_comp_assoc), tm_eq_comp_congr.
  apply tm_eq_sym, (tm_eq_trans tm_eq_fix_unfold).
  apply (tm_eq_trans tm_eq_comp_assoc), tm_eq_comp_congr.
  apply (tm_eq_trans tm_times_tm_pair), tm_eq_pair_cong; [apply tm_eq_comp_unitl|].
  apply (tm_eq_trans (tm_eq_sym tm_eq_comp_assoc)),
        (tm_eq_trans (tm_eq_comp_congl tm_eq_cons_nat_memoryless)).
  apply tm_eq_comp_assoc.
Qed.

(* Each term is equal to some Mealy-like term *)
Definition tm_eq_mealy : forall (Σ Γ : Alphabet) (t : tm Σ Γ),
  { Q : Alphabet & { δ : Σ × Q -> Q & { q₀ : Q & { p : Q -> Γ & { sp : Γ -> Q &
      (t ≡ [p]ω ∘ tm_fix q₀ [δ]ω)  * (forall b, p (sp b) = b) } } } } }%type.
  induction t; unfold tm_of_Mealy, tm_run_of_Mealy in *; simpl in *.
  - exists Γ, (fun '(a,_) => f a), (Al_pt Γ), id, id; split; [| auto].
    apply (tm_eq_trans (tm_eq_sym tm_eq_comp_unitl)), tm_eq_comp_congr,
           (tm_eq_trans (tm_fix_memoryless (Al_pt _))).
    apply tm_eq_fix_cong; solve_tm_eq_memoryless.
    destruct (w x); reflexivity.
  - destruct IHt1 as [Q1 [δ1 [q1 [p1 [sp1 [IHt1 e1]]]]]].
    destruct IHt2 as [Q2 [δ2 [q2 [p2 [sp2 [IHt2 e2]]]]]].
    repeat eexists.
    * apply (tm_eq_trans (tm_eq_comp_cong IHt1 IHt2)).
      apply (tm_eq_trans tm_eq_comp_assoc),
            (tm_eq_trans (tm_eq_comp_congr (tm_eq_sym tm_eq_comp_assoc))).
      apply (tm_eq_trans (tm_eq_comp_congr (tm_eq_comp_congl tm_fix_pre_comp))).
      apply (tm_eq_trans (tm_eq_comp_congr tm_fix_comp_fix_proj2)).
      apply (tm_eq_trans (tm_eq_sym tm_eq_comp_assoc)).
      apply tm_eq_comp_cong.
      + apply (tm_eq_sym tm_eq_memoryless_comp).
      + apply tm_eq_fix_cong.
        apply memoryless_eq_normal.
        reflexivity.
    * intros; simpl.
      instantiate (1 := fun z => (Al_pt _, sp1 z)).
      simpl; apply e1.
  - destruct IHt as [Q [δ [q [p [sp [IHt e]]]]]].
    exists (Q × Q).
    eexists; exists (sp b₀, q);
    repeat eexists.
    * apply (tm_eq_trans (tm_eq_fix_cong IHt)).
      replace (tm_fix b₀ ([p ]ω ∘ tm_fix q [δ : Σ × Γ × Q -> Q ]ω)) with
              (tm_fix (p (sp b₀)) ([p ]ω ∘ tm_fix q [δ : Σ × Γ × Q -> Q ]ω));
      [|rewrite e; reflexivity].
      apply (tm_eq_trans tm_fix_inner_memoryless_comp).
      apply (tm_eq_trans (tm_eq_comp_congr (tm_eq_fix_cong tm_fix_pre_comp))).
      apply (tm_eq_trans (tm_eq_comp_congr tm_fix_fix_proj1)).
      apply (tm_eq_trans (tm_eq_sym tm_eq_comp_assoc)).
      apply tm_eq_comp_cong.
      + apply memoryless_eq_normal; reflexivity.
      + apply tm_eq_fix_cong.
        apply memoryless_eq_normal; reflexivity.
    * intros; instantiate (1 := fun z => (sp z, sp z)).
      apply e.
Defined.

(* Now, to establish completeness, we proceed as follows:
Given t, u, if we want to check t ≡ u:
- first, reduce the problem to equality between terms of the shape
      [π]ω ∘ tm_fix a₀ [δ]ω
- given two such terms, they are extensionally equal iff
      tm_fix ((a₀,a₀'), Al_pt _) (tm_pair tm_id ([π]ω ∘ tm_proj1) ∘ δ * δ') =
      tm_fix ((a₀,a₀'), Al_pt _) (tm_pair tm_id ([π']ω ∘ tm_proj2) ∘ δ * δ')
  so the problem is reduced to the equality of fixpoints of memoryless terms
- then suppose we have two fixpoints of memoryless terms
      tm_fix a₀ δ and tm_fix a₀' δ'
  with δ : Σ × Q -> Q; if they are extensionally equal, we should have
      tm_fix a₀ δ' == tm_fix a₀' δ'
  wlog a₀ = a₀'. Define Acc(δ) the reachable states from a₀ with δ.
  If tm_fix a₀ δ == tm_fix a₀ δ', we should have Acc(δ) = Acc(δ').
  Furthermore, if ι : Acc(δ) + {a₀} -> Q is the suitable injection,
     tm_fix a₀ δ ≡ ι ∘ tm_fix r(a₀) r(δ)
  and r(δ) = r(δ'). Thus, we should be done by completeness for memoryless.
*)

(* first, let us define the * operation *)
Definition trans_prod {Σ Γ Θ : Alphabet} (t : tm (Σ × Γ) Γ) (u : tm (Σ × Θ) Θ) : tm (Σ × (Γ × Θ)) (Γ × Θ) :=
  tm_pair (t ∘ tm_times tm_id tm_proj1) (u ∘ (tm_times tm_id tm_proj2)).

Ltac tm_eq_switch := apply tm_eq_sym.
Ltac tm_eq_cong :=
  repeat (apply tm_eq_pair_cong  ||
          apply tm_eq_comp_congr ||
          apply tm_eq_comp_congl ||
          apply tm_eq_comp_cong).
Ltac tm_eq_rewrite_aux H t :=
  apply (tm_eq_trans H) ||
  match t with
   tm_pair ?t ?u => tm_eq_rewrite_aux (tm_eq_pair_cong H tm_eq_refl) t ||
                    tm_eq_rewrite_aux (tm_eq_pair_cong H tm_eq_refl) u
 | ?t ∘ ?u => tm_eq_rewrite_aux (tm_eq_comp_congl H) t ||
              tm_eq_rewrite_aux (tm_eq_comp_congr H) u
 | tm_fix _ ?t => tm_eq_rewrite_aux (tm_eq_fix_cong H) t
  end.
Ltac tm_eq_rewrite H :=
  match goal with
   [|- ?t ≡ _] => tm_eq_rewrite_aux H t
  end.

Definition tm_trans_prod_pair : forall {Σ Γ Θ : Alphabet} {t : tm (Σ × Γ) Γ} {u : tm (Σ × Θ) Θ} {a₀ b₀},
  tm_fix ((a₀, b₀) : _ × _) (trans_prod t u) ≡ tm_pair (tm_fix a₀ t) (tm_fix b₀ u).
  intros; apply tm_eq_sym, tm_eq_fix_uniq; unfold trans_prod.
  apply tm_eq_sym, (tm_eq_trans (tm_eq_sym tm_eq_pair_compr)), tm_eq_pair_cong.
  + apply tm_eq_sym, (tm_eq_trans tm_eq_fix_unfold).
    apply tm_eq_sym, (tm_eq_trans (tm_eq_comp_assoc)), tm_eq_comp_congr.
    apply (tm_eq_trans tm_times_tm_pair), tm_eq_pair_cong; [apply tm_eq_comp_unitl|].
    apply (tm_eq_trans (tm_eq_sym tm_eq_comp_assoc)),
          (tm_eq_trans (tm_eq_comp_congl tm_eq_cons_nat_memoryless)),
          (tm_eq_trans tm_eq_comp_assoc),
          tm_eq_comp_congr, tm_eq_proj1_pair.
  + apply tm_eq_sym, (tm_eq_trans tm_eq_fix_unfold).
    apply tm_eq_sym, (tm_eq_trans (tm_eq_comp_assoc)), tm_eq_comp_congr.
    apply (tm_eq_trans tm_times_tm_pair), tm_eq_pair_cong; [apply tm_eq_comp_unitl|].
    apply (tm_eq_trans (tm_eq_sym tm_eq_comp_assoc)),
          (tm_eq_trans (tm_eq_comp_congl tm_eq_cons_nat_memoryless)),
          (tm_eq_trans tm_eq_comp_assoc),
          tm_eq_comp_congr, tm_eq_proj2_pair.
Qed.

Definition tm_trans_prod_aux : forall {Σ Γ Θ Δ : Alphabet} {t : tm (Σ × Γ) Γ} {u : tm (Σ × Θ) Θ} {f : Γ × Θ -> Δ} {a₀ b₀ c₀},
   tm_fix (((a₀,b₀),c₀) : (Γ × Θ) × Δ)
      (tm_pair tm_id [f]ω ∘ trans_prod t u ∘ tm_times tm_id tm_proj1)
     ≡
   tm_pair tm_id [f]ω ∘ tm_pair (tm_fix a₀ t) (tm_fix b₀ u).
  intros.
  apply tm_eq_sym, (tm_eq_trans (tm_eq_comp_congr (tm_eq_sym tm_trans_prod_pair))), tm_eq_fix_uniq.
  apply tm_eq_sym; do 2 apply (tm_eq_trans tm_eq_comp_assoc).
  apply tm_eq_comp_congr.
  apply tm_eq_sym, (tm_eq_trans tm_eq_fix_unfold), tm_eq_comp_congr.
  apply tm_eq_sym, (tm_eq_trans tm_times_tm_pair), tm_eq_pair_cong; [apply tm_eq_comp_unitl|].
  apply (tm_eq_trans (tm_eq_sym tm_eq_comp_assoc)), (tm_eq_trans (tm_eq_comp_congl tm_eq_cons_nat_memoryless)).
  apply (tm_eq_trans (tm_eq_comp_assoc)), tm_eq_comp_congr.
  apply (tm_eq_trans (tm_eq_sym tm_eq_comp_assoc)), (tm_eq_trans (tm_eq_comp_congl tm_eq_proj1_pair)), tm_eq_comp_unitl.
Qed.

Definition tm_trans_prod_proj1 : forall {Σ Γ Θ Δ : Alphabet} {t : tm (Σ × Γ) Γ} (u : tm (Σ × Θ) Θ) {p : Γ -> Δ} {a₀} b₀ c₀,
[p]ω ∘ tm_fix a₀ t ≡
    tm_proj2 ∘ tm_fix (((a₀,b₀),c₀) : (Γ × Θ) × Δ)
      (tm_pair tm_id ([fun (z : _ × _) => p (fst z)]ω) ∘ trans_prod t u ∘ tm_times tm_id tm_proj1).
  intros.
  apply tm_eq_sym, (tm_eq_trans (tm_eq_comp_congr tm_trans_prod_aux)).
  apply (tm_eq_trans (tm_eq_sym tm_eq_comp_assoc)), (tm_eq_trans (tm_eq_comp_congl tm_eq_proj2_pair)).
  apply (tm_eq_trans (tm_eq_comp_congl tm_eq_memoryless_comp)).
  apply (tm_eq_trans tm_eq_comp_assoc), tm_eq_comp_congr, tm_eq_proj1_pair.
Qed.

Definition tm_trans_prod_proj2 : forall {Σ Γ Θ Δ : Alphabet} (t : tm (Σ × Γ) Γ) {u : tm (Σ × Θ) Θ} {p : Θ -> Δ} a₀ {b₀} c₀,
[p]ω ∘ tm_fix b₀ u ≡
    tm_proj2 ∘ tm_fix (((a₀,b₀),c₀) : (Γ × Θ) × Δ)
      (tm_pair tm_id [fun z : (_ × _) => p (snd z)]ω ∘ trans_prod t u ∘ tm_times tm_id tm_proj1).
  intros.
  apply tm_eq_sym, (tm_eq_trans (tm_eq_comp_congr tm_trans_prod_aux)).
  apply (tm_eq_trans (tm_eq_sym tm_eq_comp_assoc)), (tm_eq_trans (tm_eq_comp_congl tm_eq_proj2_pair)).
  apply (tm_eq_trans (tm_eq_comp_congl tm_eq_memoryless_comp)).
  apply (tm_eq_trans tm_eq_comp_assoc), tm_eq_comp_congr, tm_eq_proj2_pair.
Qed.

Lemma trans_prod_fun_eq : forall {Σ Γ Θ Δ : Alphabet} (t : tm (Σ × Γ) Γ) (u : tm (Σ × Θ) Θ) (p : Γ -> Δ) (q : Θ -> Δ) (a₀ : Γ) (b₀ : Θ) (c₀ : Δ) (w : nat -> Σ) (n : nat),
  fun_of_tm ([p]ω ∘ tm_fix a₀ t) w n = fun_of_tm ([q]ω ∘ tm_fix b₀ u) w n
  ->
    fun_of_tm (tm_fix (((a₀,b₀),c₀) : (Γ × Θ) × Δ)
      (tm_pair tm_id ([fun (z : _ × _) => p (fst z)]ω) ∘ trans_prod t u ∘ tm_times tm_id tm_proj1)) w n
=
    fun_of_tm (tm_fix (((a₀,b₀),c₀) : (Γ × Θ) × Δ)
      (tm_pair tm_id [fun z : (Γ × Θ) => q (snd z)]ω ∘
                      (trans_prod t u : tm (Σ × (Γ × Θ)) (Γ × Θ)) ∘ tm_times tm_id tm_proj1)) w n. 
  intros.
  do 2 rewrite (soundness _ _ _ _ tm_trans_prod_aux).
  do 2 rewrite tm_comp_fun_eq, tm_pair_fun_eq ; unfold pairfun; apply f_equal.
  do 2 rewrite tm_comp_fun_eq in H;
  do 2 rewrite tm_memoryless_fun_eq; simpl.
  rewrite tm_pair_fun_eq; unfold pairfun; simpl.
  do 2 rewrite tm_memoryless_fun_eq in H.
  exact H.
Qed.

(* Now, it remains to show that for every f, g : Σ × Γ -> Γ and b₀ : Γ
  fix b₀ [f]ω == fix b₀ [g]ω -> fix b₀ [f]ω ≡ fix b₀ [g]ω

Note that we do not necessarily have f == g. However, we should be able to
find suitable restrictions of f and g which agrees.

technically speaking, there is ι : Γ' -> Γ, c₀ : Γ' and δ' : Σ × Γ' -> Γ' st
ι c₀ = b₀
ι^-1 ∘ f ∘ pairfun id ι = ι^-1 ∘ g ∘ pairfun id ι
and fix b₀ f ≡ ι ∘ fix c₀ δ'
*)

Definition reachable_trans_pred {Σ Γ : Alphabet} (f : Σ × Γ -> Γ) p (b : Γ) : bool :=
  p b || exf (fun a =>exf (fun b' => p b' && if FT_deceq Γ (f (a, b')) b then true else false)).

Lemma reachable_trans_pred_infl {Σ Γ : Alphabet} (f : Σ × Γ -> Γ) : Kleenefin.inflationary _ (reachable_trans_pred f).
  unfold reachable_trans_pred; intros ? ? e; rewrite e; reflexivity.
Qed.
  
Definition Kleene_reachable {Σ Γ : Alphabet} (b₀ : Γ) (f : Σ × Γ -> Γ) :
  { n : nat |  reachable_trans_pred f (iter (reachable_trans_pred f) n (fun b => if FT_deceq _ b b₀ then true else false)) == iter (reachable_trans_pred f) n (fun b => if FT_deceq _ b b₀ then true else false)}.
  apply Kleene.
  + intros ? ? ? ?.
    apply orb_mono; auto.
    do 2 (apply exf_mono; intro).
    apply andb_mono; auto.
  + unfold reachable_trans_pred; intros ? ? ? ?.
    rewrite H; apply f_equal.
    do 2 (apply exf_ext; intro).
    rewrite H; reflexivity.
  + apply reachable_trans_pred_infl.
Defined.

Definition reachable {Σ Γ : Alphabet} (b₀ : Γ) (f : Σ × Γ -> Γ) :=
  iter (reachable_trans_pred f) (proj1_sig (Kleene_reachable b₀ f)) (fun b => if FT_deceq _ b b₀ then true else false).

Lemma reachable_init {Σ Γ : Alphabet} (b₀ : Γ) {f : Σ × Γ -> Γ} : reachable b₀ f b₀.
  unfold reachable.  
  apply Kleenefin.infl_iter.
  + apply reachable_trans_pred_infl.
  + destruct (FT_deceq _ b₀ b₀); [reflexivity|exfalso; congruence].
Qed.

Lemma reachable_step {Σ Γ : Alphabet} (b₀ : Γ) {f : Σ × Γ -> Γ} b a : reachable b₀ f b -> reachable b₀ f (f (a,b)).
  unfold reachable.  
  destruct (Kleene_reachable b₀ f); simpl in *.
  remember (iter (reachable_trans_pred f) x (fun b0 : Γ => if FT_deceq Γ b0 b₀ then true else false)) as p.
  clear Heqp.
  intro; rewrite <-e; clear e.
  unfold reachable_trans_pred.
  destruct (p (f (a,b))); simpl;[reflexivity|].
  apply exf_exists; exists a;
  apply exf_exists; exists b.
  rewrite H.
  destruct (FT_deceq Γ (f (a,b)) (f (a,b))); [reflexivity|exfalso; auto].
Qed.

Inductive reachable_prop {Σ Γ : Alphabet} (b₀ : Γ) (f : Σ × Γ -> Γ) : Γ -> Prop :=
  reachable_prop_init : reachable_prop b₀ f b₀
| reachable_prop_step : forall a b, reachable_prop b₀ f b -> reachable_prop b₀ f (f (a,b)).

Lemma reachable_prop_reachable {Σ Γ : Alphabet} (b₀ b : Γ) (f : Σ × Γ -> Γ) : reachable b₀ f b <-> reachable_prop b₀ f b.
  split.
  + unfold reachable. revert b.
    induction (proj1_sig (Kleene_reachable b₀ f)); simpl; intros.
    - destruct (FT_deceq Γ b b₀); [subst; constructor|inversion H].
    - unfold reachable_trans_pred in *.
      apply orb_prop in H.      
      destruct H; auto.
      do 2 (apply exf_exists in H; destruct H as [? H]).
      apply andb_prop in H.
      destruct H.
      destruct (FT_deceq Γ (f (x,x0)) b); [subst|inversion H0].
      constructor; auto.
  + intro p; induction p.
    - apply reachable_init.
    - apply reachable_step; assumption.
Qed.

Inductive reachable_prop' {Σ Γ : Alphabet} (b₀ : Γ) (f : Σ × Γ -> Γ) : Γ -> Prop :=
  reachable_prop'_init : reachable_prop' b₀ f b₀
| reachable_prop'_step : forall a b, reachable_prop' (f (a,b₀)) f b -> reachable_prop' b₀ f b.

Lemma reachable_prop_prop' {Σ Γ : Alphabet} (b₀ b : Γ) (f : Σ × Γ -> Γ) : reachable_prop b₀ f b <-> reachable_prop' b₀ f b.
  split; (intro p; induction p; [constructor|]).
  + clear p; revert a; induction IHp.
    - intro; econstructor; constructor.
    - intro.
      econstructor; auto.
  + clear p.
    remember (f (a,b₀)) as b'; revert a Heqb'.
    induction IHp; intros; subst.
    - do 2 constructor.
    - constructor.
      eapply IHIHp; reflexivity.
Qed.

Fixpoint list_of_seq {A : Type} (s : nat -> A) (n : nat) :=
  match n with
   0 => nil
 | S k => s 0 :: list_of_seq (fun n => s (S n)) k
  end.

Lemma fun_of_tm_fix_S : forall {Σ Γ : Alphabet} (f : Σ × Γ -> Γ) b₀,
 forall a w x, fun_of_tm (tm_fix b₀ [f]ω) (fun n => match n with 0 => a | S k => w k end) (S x) = fun_of_tm (tm_fix (f (a,b₀)) [f]ω) w x.
  intros; unfold fun_of_tm, fun_of_Mealy in *.
  apply f_equal with (f:= fst).
  induction x; [reflexivity|].
  simpl in *.  
  destruct (run_of_Mealy (Mealy_fix b₀ (Mealy_memoryless (f : _ × _ ->_)))
               (fun n : nat => match n with
                               | 0 => a
                               | S k => w k
                               end) x).
  destruct (run_of_Mealy (Mealy_fix (f (a, b₀)) (Mealy_memoryless (f : _ × _ -> _))) w x).
  destruct f1, f3; simpl in *.
   congruence.
Qed.

Lemma fun_of_tm_fix_shift_eq : forall {Σ Γ : Alphabet} (f g : Σ × Γ -> Γ) b₀,
  (forall w, fun_of_tm (tm_fix b₀ [f]ω) w == fun_of_tm (tm_fix b₀ [g]ω) w) ->
  (forall a, f (a, b₀) = g (a,b₀) /\ (forall w, fun_of_tm (tm_fix (f (a,b₀)) [f]ω) w == fun_of_tm (tm_fix (g (a,b₀)) [g]ω) w)).
  intros.
  split.
  + apply (H (fun _ => a) 0).
  + intros ? ?.
    specialize (H (fun n => match n with 0 => a | S k => w k end) (S x)).
    do 2 rewrite fun_of_tm_fix_S in H; assumption.
Qed.

Lemma same_reachable_aux : forall {Σ Γ : Alphabet} (b₀ : Γ) (f g : Σ × Γ -> Γ),
  (forall w, fun_of_tm (tm_fix b₀ [f]ω) w == fun_of_tm (tm_fix b₀ [g]ω) w)
 -> forall b, reachable b₀ f b -> reachable b₀ g b && allf (fun a => if FT_deceq _ (f (a,b)) (g (a,b)) then true else false).
  intros.
  apply reachable_prop_reachable in H0.
  apply reachable_prop_prop' in H0.
  induction H0.  
  + rewrite reachable_init; apply allf_forall.
    intro x; destruct (FT_deceq _ (f (x, b₀)) (g (x, b₀))); [reflexivity|].
    eapply fun_of_tm_fix_shift_eq in H; exfalso; destruct H; eauto.
  + assert (H' := H).
    apply fun_of_tm_fix_shift_eq with (a0 := a) in H; destruct H.
    exploit IHreachable_prop'; [intros ? ?; rewrite H1, H; reflexivity|].
    apply andb_prop in IHreachable_prop'.
    destruct IHreachable_prop'.
    apply andb_true_iff; split.
    - apply reachable_prop_reachable, reachable_prop_prop'.
      econstructor. rewrite <-H.
       apply reachable_prop_prop', reachable_prop_reachable.
       assumption.
    - assumption.
Qed.

Lemma same_reachable_set : forall {Σ Γ : Alphabet} (b₀ : Γ) (f g : Σ × Γ -> Γ),
  (forall w, fun_of_tm (tm_fix b₀ [f]ω) w == fun_of_tm (tm_fix b₀ [g]ω) w)
 -> forall b, reachable b₀ f b -> reachable b₀ g b.
  intros.
  specialize (same_reachable_aux b₀ f g H b H0); intro  H'.
     apply andb_prop in H'; destruct H';assumption.
Qed.

Lemma reachable_eq : forall {Σ Γ : Alphabet} (b₀ : Γ) (f g : Σ × Γ -> Γ),
  (forall w, fun_of_tm (tm_fix b₀ [f]ω) w == fun_of_tm (tm_fix b₀ [g]ω) w)
 -> forall b, reachable b₀ f b -> forall a : Σ,  f (a,b) = g (a,b).
  intros.
  specialize (same_reachable_aux b₀ f g H b H0); intro  H'.
     apply andb_prop in H'; destruct H'.
     specialize (@allf_forall _ (fun a : Σ => if FT_deceq Γ (f (a, b)) (g (a, b)) then true else false)).
     intros [? _].
     specialize (H3 H2 a).
     destruct (FT_deceq _ (f (a,b)) (g (a,b)));[assumption|inversion H3].
Qed.

Lemma tm_reduce : forall {Σ Γ Γ' : Alphabet} (c₀ : Γ') (f : Σ × Γ -> Γ)
(s : Γ' -> Γ) (r : Γ -> Γ'), (forall x : Γ', r (s x) = x) -> (forall a b, f (a,s b) = s (r (f (a, s b)))) -> [s]ω ∘ tm_fix c₀ ([fun (z : _ × _) => let (a,b) := z in r (f (a,s b))]ω) ≡ tm_fix (s c₀) [f]ω.
  intros; apply tm_eq_fix_uniq.
  apply (tm_eq_trans (tm_eq_comp_congr tm_eq_fix_unfold)).
  apply (tm_eq_trans (tm_eq_sym tm_eq_comp_assoc)).
  apply (tm_eq_trans (tm_eq_comp_congl (tm_eq_sym tm_eq_memoryless_comp))).
  apply tm_eq_sym.
  apply (tm_eq_trans (tm_eq_comp_congr (tm_eq_pair_cong tm_eq_refl (tm_eq_sym tm_eq_comp_assoc)))).
  apply (tm_eq_trans (tm_eq_comp_congr (tm_eq_pair_cong (tm_eq_sym tm_eq_comp_unitl) (tm_eq_comp_congl (tm_eq_sym tm_eq_cons_nat_memoryless))))).
  apply (tm_eq_trans (tm_eq_comp_congr (tm_eq_pair_cong tm_eq_refl tm_eq_comp_assoc))).
  apply (tm_eq_trans (tm_eq_comp_congr (tm_eq_sym tm_times_tm_pair))).
  apply (tm_eq_trans (tm_eq_sym tm_eq_comp_assoc)).
  apply tm_eq_comp_congl.
  solve_tm_eq_memoryless.
  rewrite tm_pair_fun_eq; unfold pairfun;simpl.
  repeat rewrite tm_comp_fun_eq, tm_memoryless_fun_eq.
  rewrite tm_proj1_fun_eq, tm_proj2_fun_eq.
  destruct (w x); auto.
Qed.

Definition complete : forall {Σ Γ : Alphabet} (t u : tm Σ Γ), (forall w, fun_of_tm t w == fun_of_tm u w) -> t ≡ u.
  intros.
  specialize (tm_eq_mealy _ _ t); specialize (tm_eq_mealy _ _ u).
  intros [Qu [δu [qu₀ [pu [spu [Hu eu]]]]]]
         [Qt [δt [qt₀ [pt [spt [Ht et]]]]]].
  apply (tm_eq_trans Ht), tm_eq_sym, (tm_eq_trans Hu), tm_eq_sym.
  specialize (soundness _ _ _ _ Hu);
  specialize (soundness _ _ _ _ Ht).
  intros.
  assert (forall w, fun_of_tm ([pt]ω ∘ tm_fix qt₀ [δt]ω) w == 
                    fun_of_tm ([pu]ω ∘ tm_fix qu₀ [δu]ω) w) by congruence.
  clear H H0 H1 t u Hu Ht.
  assert (H3 := fun (w : nat -> Σ) n => trans_prod_fun_eq _ _ _ _ _ _ (Al_pt _) w n (H2 w n)).
  clear H2.
  apply (tm_eq_trans (tm_trans_prod_proj1 [δu]ω qu₀ (Al_pt _))).
  apply tm_eq_sym, (tm_eq_trans (tm_trans_prod_proj2 [δt]ω qt₀ (Al_pt _))).
  apply tm_eq_comp_congr.
  specialize (@memoryless_eq_normal (Σ × ((Qt × Qu) × Γ)) _ (tm_pair tm_id [fun z : Qt × Qu => pu (snd z) ]ω ∘ trans_prod [δt ]ω [δu ]ω ∘ tm_times tm_id tm_proj1)).
  intros.
  exploit X; [reflexivity|].
  match type of X with
   _ ≡ [?t]ω => remember t as m end.
  apply (tm_eq_trans (tm_eq_fix_cong X)).
  specialize (@memoryless_eq_normal (Σ × ((Qt × Qu) × Γ)) _ (tm_pair tm_id [fun z : Qt × Qu => pt (fst z) ]ω ∘ trans_prod [δt ]ω [δu ]ω ∘ tm_times tm_id tm_proj1)); intros Y; exploit Y; [reflexivity|].
  match type of Y with
   _ ≡ [?t]ω => remember t as n end.
  apply tm_eq_sym, (tm_eq_trans (tm_eq_fix_cong Y)).
  assert (forall w, fun_of_tm (tm_fix (((qt₀, qu₀), Al_pt _) : (_ × _) × _) [m]ω) w ==
                    fun_of_tm (tm_fix (((qt₀, qu₀), Al_pt _) : (_ × _) × _) [n]ω) w).
  {
    intros w x; specialize (H3 w x).
    specialize (soundness _ _ _ _ X); specialize (soundness _ _ _ _ Y); clear X Y; intros Y X.
    rewrite (tm_fix_ext X) in H3.
    rewrite (tm_fix_ext Y) in H3; symmetry.
    apply H3.
  }
  clear pt spt pu spu et eu X Y Heqm Heqn H3.
  remember (qt₀, qu₀, Al_pt Γ) as q₀.
  clear Heqq₀ qt₀ qu₀ δt δu.
  specialize (same_reachable_set (q₀ : (_ × _) × _) m n H); intros.
  specialize (same_reachable_set (q₀ : (_ × _) × _) n m (fun w n => eq_sym (H w n))); intros.
  pose (Q := (Qt × Qu) × Γ).
  pose (Q' := (Al_sub (reachable (q₀ : (_ × _) × _) m) q₀ (reachable_init (q₀ : (_ × _) × _)))).
  pose (n':= fun (z : Σ × Q') => let (a,b) return Q' := z in sub_retract (n (a, sub_section b))).
  specialize (tm_reduce (sub_retract (q₀ : Q)) n sub_section sub_retract (sub_section_retraction _ (reachable (q₀ : Q) n) q₀ (reachable_init _))); intros.
  specialize (tm_reduce (sub_retract (q₀ : Q)) m sub_section sub_retract (sub_section_retraction _ (reachable (q₀ : Q) n) q₀ (reachable_init _))); intros Y.
  exploit X.
  {
    intros.
    symmetry; apply sub_retraction_section.
    apply reachable_step.
    apply sub_prop.
  }
  exploit Y.
  {
    intros.
    symmetry; apply sub_retraction_section.
    assert (reachable (q₀ : Q) m (sub_section b)).
    {
      apply H1, sub_prop.
    }
    assert (m (a, sub_section b) = n (a, sub_section b)).
    {
      apply (reachable_eq _ _ _ H), H2.
    }
    apply H0, reachable_step, H2.
  }
  rewrite <-(sub_retraction_section _ (reachable (q₀ : Q) n) q₀ (reachable_init _) (q₀ : Q)); [|apply reachable_init].
  clear H0 H1.
  apply tm_eq_sym in X; apply tm_eq_sym in Y.
  apply (tm_eq_trans X), tm_eq_sym, (tm_eq_trans Y).
  apply tm_eq_comp_congr, tm_eq_fix_cong, tm_eq_memoryless.
  intros [? ?].
  specialize (@reachable_eq _ _ _ _ _ (fun w n => eq_sym (H w n))); intros.
  apply f_equal.
  symmetry; apply H0, sub_prop.
Defined.
