Require Import Bool.
Require Import List.
Require Import Lia.

(* Some preliminary definitions.
   Mostly has to do with finite types that we use as alphabets. *)

Notation "f == g" := (forall x, f x = g x) (at level 53).
Ltac exploit H :=
  match type of H with
   ?A -> _ => let x := fresh in assert A as x;[| specialize (H x); clear x]
  end.

Definition pairfun {A B C : Type} (f : A -> B) (g : A -> C) (a : A) : B * C := (f a, g a).

Coercion Prop_of_bool (b : bool) := b = true.

Definition EqDec (T : Type) := forall (x y : T), {x = y} + {~x = y}.

Record FinType :=
{
  FT_carrier :> Type;
  FT_deceq : EqDec FT_carrier;
  FT_enum : list FT_carrier;
  FT_enum_all : forall x, In x FT_enum;
}.

Record Alphabet :=
{
  Al_carrier :> FinType;
  Al_pt : Al_carrier;
}.

Definition Fin_unit : FinType :=
  {| FT_carrier := unit;
     FT_deceq := fun x => match x with tt => fun y => match y with tt => left eq_refl end end;
     FT_enum := tt :: nil;
     FT_enum_all := fun x => match x with tt => or_introl eq_refl end;
  |}.

Definition Al_unit : Alphabet :=
  {| Al_carrier := Fin_unit;
     Al_pt := tt
  |}.

Definition Fin_prod (Σ Γ : FinType) : FinType.
  unshelve esplit.
  - exact (Σ * Γ)%type.
  - exact (list_prod (FT_enum Σ) (FT_enum Γ)).
  - intros [a b] [a' b'];
    destruct (FT_deceq Σ a a'), (FT_deceq Γ b b'); subst;
    try (right; congruence); left; reflexivity.
  - intros [a b]; simpl; apply in_prod; apply FT_enum_all.
Defined.

Definition Al_prod (Σ Γ : Alphabet) : Alphabet :=
  {| Al_carrier := Fin_prod Σ Γ;
     Al_pt := (Al_pt Σ, Al_pt Γ);
  |}.

Definition Fin_bool : FinType.
  unshelve esplit.
  - exact bool.
  - exact (true :: false :: nil).
  - intros; unfold EqDec; decide equality.
  - intros [|]; intuition.
Defined.

Definition Al_bool : Alphabet :=
  {| Al_carrier := Fin_bool;
     Al_pt := true;
  |}.

Notation "Σ × Γ" := (Al_prod Σ Γ) (at level 13, left associativity).

Definition UIP (T : Type) := forall (x y : T) (p q : x = y), p = q.

Lemma hedberg : forall T, EqDec T -> UIP T.
  intros ? ?.
  pose (fun (x y : T) (e : x = y) => match X x y with left e' => e' | right p => match (p e) with end end).
  assert (forall x y p q, e x y p = e x y q).
  {
    intros; subst e; simpl; destruct (X x y);[reflexivity|].
    destruct (n p).
  }
  pose (s := @eq_sym T).
  pose (t := @eq_trans T).
  intros x y.
  assert (forall p : x = y, p = t _ _ _ (s _ _ (e _ _ eq_refl)) (e _ _ p)).
  {
    intro p; subst t s; unfold eq_sym, eq_trans.
    destruct p.
    subst e; simpl.
    destruct (X x x). 
    - destruct e; reflexivity.
    - destruct (n eq_refl).
  }
  intros.
  rewrite (H0 p), (H0 q), (H _ _ p q); reflexivity.
Defined.

Scheme J := Induction for eq Sort Type.

Lemma FinTypeUIP : forall C : FinType, UIP C.
  intros [? ? ?]; simpl in *; apply hedberg; assumption.
Qed.

Fixpoint filter_inl {A B : Type} (l : list (A + B)) : list A :=
  match l with
   nil => nil
 | inl h :: t => h :: filter_inl t
 | inr _ :: t => filter_inl t
  end.

Lemma filter_inl_in : forall {A B : Type} (l : list (A + B)) a, In (inl a) l -> In a (filter_inl l).
  induction l; simpl; intros; auto.
  destruct H.
  - subst.
    left; reflexivity.
  - destruct a;[right|]; auto.
Qed.

Definition sig_incl {A : Type} {P Q : A -> Prop} (i : forall a, P a -> Q a) : sig P -> sig Q := fun z => match z with exist _ w p => exist _ w (i _ p) end.

(*
Ideally, we would like to use the following definition:

Definition bool_split {A : Type} (p : A -> bool) : A -> { a : A | p a } + { a : A | p a = false } :=
  fun z => match p z as x return p z = x -> _ with
            true => fun e => inl (exist _ z e)
          | false => fun e => inr (exist _ z e)
           end eq_refl.

But proofs are shorter with the following which work over the larger A * (A -> bool).
*)

Definition bool_split0 {A : Type} : A * (A -> bool) ->
  { z : A * (A -> bool) & (snd z (fst z)) + (snd z (fst z) = false) }%type :=
  fun '(a,p) => existT _ (a,p)
                 (match p a as x return p a = x -> _ with
                   true => inl
                 | false => inr end eq_refl).

Definition bool_split1 {A : Type} : { z : A * (A -> bool) & (snd z (fst z)) + (snd z (fst z) = false) }%type ->
{ p : (A -> bool) & { a : A | p a } + { a : A | p a = false } }%type :=
  fun z => match z with
            existT _ (a,p) z => existT _ p (match z with
                                             inl q => inl (exist _ a q)
                                           | inr q => inr (exist _ a q)
                                            end)
           end.

(*Lemma bool_pred_split_eq {A : Type} (x : A) p :
   bool_pred_split p x = projT2 (bool_pred_split1 (bool_pred_split0 (x,p))).*)

Definition bool_split {A : Type} (p : A -> bool) (x : A) : {a | p a} + {a | p a = false} := projT2 (bool_split1 (bool_split0 (x,p))).

Definition bool_unsplit {A : Type} (p : A -> bool) : { a : A | p a } + { a : A | p a = false } -> A:=
  fun z => match z with
            inl (exist _ x _) => x
          | inr (exist _ x _) => x
           end.

Lemma bool_unsplit_split {A : Type} (p : A -> bool) : forall x, bool_unsplit p (bool_split p x) = x.
  intros.
  unfold bool_split, bool_unsplit, bool_split1, bool_split0.
  simpl.
  destruct ( (if p x as x0 return (p x = x0 -> (p x = true) + (p x = false)) then inl else inr)
        eq_refl
); reflexivity.
Qed.


Lemma bool_split_unsplit {A : Type} (p : A -> bool) : forall x, bool_unsplit p (bool_split p x) = x.
  intros.
  unfold bool_split, bool_unsplit, bool_split1, bool_split0.
  simpl.
  destruct ( (if p x as x0 return (p x = x0 -> (p x = true) + (p x = false)) then inl else inr)
        eq_refl
); reflexivity.
Qed.

Definition filter_refine_trivial {A : Type} (p : A -> bool) (l : list A) : list ({ a : A | p a } + {a : A | p a = false}) := map (bool_split p) l.

Definition filter_refine {A : Type} (p : A -> bool) (l : list A) : list { a : A | p a } :=
  filter_inl (filter_refine_trivial p l).

(* Fixpoint filter_refine {A : Type} (p : A -> bool) (l : list A) : list { a : A | p a } :=
  match l with
   nil => nil
|  h :: t => match p h as x return p h = x -> _ with
              true => fun e => (exist p h e) :: filter_refine p t
            | false => fun e => filter_refine p t
             end eq_refl
  end. *)

Lemma exist_proj1eq : forall {A : Type} {p : A -> bool} {a b : A}, a = b -> forall z z', exist p a z = exist p b z'.
  intros; subst; apply f_equal, (FinTypeUIP Fin_bool).
Qed.

Lemma filter_refine_in : forall {A : Type} (p : A -> bool) (l : list A) a, In a l -> forall z : p a, In (exist _ a z) (filter_refine p l).
  induction l; unfold filter_refine, filter_refine_trivial in *; simpl; intros; auto.
  destruct H; subst.
  + case_eq (bool_split p a0); intros [? ?] ?.
    - left.
      apply exist_proj1eq.      
      apply (f_equal (bool_unsplit p)) in H.
      rewrite bool_split_unsplit in H.
      compute in H; auto.
    - apply (f_equal (bool_unsplit p)) in H.
      rewrite bool_split_unsplit in H.
      compute in H; subst.
      exfalso; congruence.
  + destruct (bool_split p a);[right|]; apply IHl; assumption.
Qed.

Definition Fin_sub {A : FinType} (p : A -> bool) : FinType.
  unshelve esplit.
  - exact { a : A | p a }.
  - exact (filter_refine p (FT_enum A)).
  - intros [? ?] [? ?].    
    destruct (FT_deceq A x x0).
    + left; subst; apply f_equal.
      apply (FinTypeUIP Fin_bool).
    + right; intro.
      inversion H; auto.
  - intros [? ?].
    apply filter_refine_in, FT_enum_all.
Defined.

Definition Al_sub {Σ : Alphabet} (p : Σ -> bool) (a : Σ) (e : p a) : Alphabet :=
 {| Al_carrier := Fin_sub p;
    Al_pt := exist p a e;
 |}.

Definition sub_section {Σ : Alphabet} {p : Σ -> bool} {a : Σ} {e : p a} : Al_sub p a e -> Σ := fun z => let (x, _) := z in x.

Definition Al_select {Σ : Alphabet} (p : Σ -> bool) : Σ :=
  match filter p (FT_enum Σ) with
   h :: _ => h
 | nil => Al_pt _
  end.

Lemma Al_select_sound {Σ : Alphabet} (p : Σ -> bool) : forall x, p x -> p (Al_select p).
  intros.
  unfold Al_select.
  case_eq (filter p (FT_enum Σ)).
  + intros; exfalso.
    assert (In x (filter p (FT_enum Σ))).
    {
       apply filter_In; split; auto.
       apply FT_enum_all.
    }
    rewrite H0 in H1; apply H1.
  + intros.
    assert (In f (filter p (FT_enum Σ))).
    {
       rewrite H0; left; reflexivity.
    }
    destruct (let (z,_) := (filter_In _ _ _) in z H1); assumption.
Qed.

Definition sub_retract {Σ : Alphabet} {p : Σ -> bool} {a : Σ} {e : p a} : Σ -> Al_sub p a e := fun x => Al_select (fun (z : Al_sub p a e) => if FT_deceq _ (sub_section z) x then true else false).

Lemma sub_section_retraction : forall (Σ : Alphabet) (p : Σ -> bool) a e (x : Al_sub p a e), sub_retract (sub_section x) = x.
  intros.
  destruct x; unfold sub_retract, sub_section in *.
  specialize (@Al_select_sound (Al_sub p a e) (fun z => if FT_deceq _ (sub_section z) x then true else false) (exist _ x p0)); intros.
  exploit H.
  + case_eq (FT_deceq Σ (@sub_section Σ p a e (exist p x p0)) x); intros; [reflexivity|].
    simpl in *.
    exfalso; apply n; auto.
  + simpl in *.
    destruct (FT_deceq Σ
         (@sub_section Σ p a e
            (@Al_select (@Al_sub Σ p a e)
               (fun z : {a : Σ | p a} =>
                if FT_deceq Σ (@sub_section Σ p a e z) x then true else false))) x).
    - unfold sub_section in *.
      Set Printing Implicit.
      destruct (@Al_select (@Al_sub Σ p a e)
    (fun z : {a0 : Σ | p a0} =>
     if FT_deceq Σ (let (x0, _) := z in x0) x then true else false)).
      subst; apply f_equal, hedberg, (FT_deceq Fin_bool).
    - inversion H.
Qed.

Lemma sub_retraction_section : forall (Σ : Alphabet) (p : Σ -> bool) a e (x : Σ), p x -> @sub_section Σ p a e (sub_retract x) = x.
  intros.
  unfold sub_retract, sub_section in *.
  specialize (@Al_select_sound (@Al_sub Σ p a e)
       (fun z : @Al_sub Σ p a e =>
        if FT_deceq Σ (@sub_section Σ p a e z) x then true else false) (exist _ x H)).
  intros X; exploit X.
  {
     destruct (FT_deceq Σ (@sub_section Σ p a e (@exist Σ (fun a0 : Σ => p a0) x H)) x);
     [reflexivity|exfalso]. apply n; reflexivity.
  }
  simpl in X.
  destruct (FT_deceq Σ
         (@sub_section Σ p a e
            (@Al_select (@Al_sub Σ p a e)
               (fun z : {a : Σ | p a} =>
                if FT_deceq Σ (@sub_section Σ p a e z) x then true else false))) x).
  + unfold sub_section in *.
    compute in *; rewrite e0; reflexivity.
  + inversion X.
Qed.

Lemma sub_prop : forall (Σ : Alphabet) (p : Σ -> bool) a e (x : Al_sub p a e), p (sub_section x).
  intros.
  destruct x.
  exact p0.
Qed.


Fixpoint iter {A : Type} (f : A -> A) (n : nat) a :=
    match n with
     0 => a
   | S k => f (iter f k a)
    end.

Lemma iter_plus {A : Type} (f : A -> A) (n k : nat) a : iter f (n + k) a = iter f n (iter f k a).
  induction n; simpl; eauto.
  rewrite IHn; reflexivity.
Qed.

Definition exf {Σ : FinType} (p : Σ -> bool) := existsb p (FT_enum Σ).

Lemma exf_exists : forall {Σ: FinType} {p : Σ -> bool}, exf p <-> exists x, p x.
  intros.
  split; intro; unfold exf in *.
  + apply existsb_exists in H; destruct H as [? [? ?]]; eauto.
  + apply existsb_exists.
    destruct H as [w ?]; exists w; split; auto.
    apply FT_enum_all.
Qed.

Definition allf {Σ : FinType} (p : Σ -> bool) := negb (exf (fun x => negb (p x))).

Lemma allf_forall : forall {Σ: FinType} {p : Σ -> bool}, allf p <-> forall x, p x.
  intros;  split; intros.
  + case_eq (p x); intro; [reflexivity|].
    specialize (negb_true_iff (p x)).
    intros [_ ?].
    apply H1 in H0. 
    assert (exf (fun x => negb (p x))) by (apply exf_exists; eauto).
    unfold allf in H; rewrite H2 in H.
    inversion H.
  + apply negb_true_iff.
    case_eq (exf (fun x => (negb (p x)))); intro; try reflexivity.
    apply exf_exists in H0.
    destruct H0.
    rewrite H in H0; inversion H0.
Qed.

(* A fixpoint theorem over finite sets for decidable predicates. *)
Module Kleenefin.
Section Kleenefin.

  Variable Σ : Type.
  Variable deceq : EqDec Σ.

  Fixpoint in_b (a : Σ) (l : list Σ) : bool :=
    match l with
     nil => false
   | h :: t => if deceq a h then true else in_b a t
    end.

  Definition mono (F : (Σ -> bool) -> Σ -> bool) :=
    forall (p q : Σ -> bool), (forall x, p x -> q x) -> forall x, F p x -> F q x.

  Definition bounded (l : list Σ) (F : (Σ -> bool) -> Σ -> bool) :=
    forall (p : _ -> bool), (forall x, p x -> in_b x l) -> forall x, F p x -> in_b x l.

  Lemma iter_mono : forall {F}, mono F -> forall {q}, (forall x, F q x -> q x) -> forall n x, iter F n q x -> q x.
    induction n; auto.
    simpl; intros.
    eapply H0, H, H1; auto.
  Qed.

  Definition ext (F : (Σ -> bool) -> Σ -> bool) := forall p q, p == q -> F p == F q.

  Definition inflationary (F : (Σ -> bool) -> Σ -> bool) := forall (p : _ -> bool) x, p x -> F p x.


  Lemma infl_iter : forall F, inflationary F -> forall n, inflationary (iter F n).
    induction n; intros ? ? ?; simpl; eauto.
  Qed.

  Lemma remove_length : forall x l, length (remove deceq x l) <= length l.
    induction l; simpl; auto.
    destruct (deceq x a); simpl; auto.
    lia.
  Qed.

  Lemma remove_in_b_length : forall x l, in_b x l -> length (remove deceq x l) < length l.
    intros; induction l;simpl in *;intros; [exfalso; congruence|].
    destruct (deceq x a);subst.
    + specialize (remove_length a l); lia.
    + simpl. specialize (IHl H); lia.
  Qed.

  Lemma remove_in_b : forall l x y, in_b x l -> x <> y -> in_b x (remove deceq y l).
    induction l; intros; auto.
    simpl in *.
    destruct (deceq x a), (deceq y a); subst; simpl in *; try (exfalso; congruence); eauto.
    + destruct (deceq a a); [reflexivity| exfalso; congruence].
    + destruct (deceq x a); [reflexivity| apply IHl; auto].
  Qed.

  Lemma remove_not_in_b : forall l x, in_b x (remove deceq x l) = false.
    induction l; intros; simpl; auto.
    destruct (deceq x a); simpl; auto.
    destruct (deceq x a); simpl; auto; exfalso; auto.
  Qed.

  Lemma remove_in_b' : forall l x y, in_b x (remove deceq y l) -> in_b x l.
    induction l; simpl in *; intros; eauto.
    destruct (deceq x a), (deceq y a); simpl in *; subst; try reflexivity; eauto.
    destruct (deceq x a); subst; try (exfalso; congruence); eauto.
  Qed.


  Lemma iter_bound : forall l F, bounded l F -> forall p : _ -> bool, (forall x, p x -> in_b x l) -> forall n x, iter F n p x -> in_b x l.
    induction n; simpl in *; intros; eauto.
  Qed.

  Theorem Kleene : forall l (p: _ -> bool) F, (forall z, p z -> in_b z l) -> inflationary F -> bounded l F -> mono F -> ext F ->
    F (iter F (length l) p) == iter F (length l) p.
    intros l p; remember (length l) as n.
    revert l Heqn. induction (Wf_nat.lt_wf n); intros; simpl in *.
    rename x into n, x0 into a.
    destruct n.
    - destruct l; simpl in *; auto.
      + assert (p a = false) by (case_eq (p a); intro; [symmetry; eapply H1; eauto|reflexivity]).
        rewrite H6; case_eq (F p a); [|intro; reflexivity].
        intro.
        rewrite H3; eauto.
      + inversion Heqn.
    - match goal with
       [|- ?t = ?u] => case_eq t; case_eq u; intros E E'; try reflexivity; exfalso
      end.
      + pose (G := fun p y => if deceq a y then p y else F p y).
        assert (mono G).
        {
          subst G; intros ? ? ? ? ?.
          destruct (deceq a x); eauto.
        }
        assert (inflationary G).
        {
           subst G; intros ? ? ?.
           destruct (deceq a x); eauto.
        }
        assert (bounded (remove deceq a l) G).
        {
           intros ? ? ?; subst G; simpl in *.
           destruct (deceq a x); intro; subst.
           + specialize (H8 _ H9). rewrite remove_not_in_b in H8; inversion H8.
           + apply remove_in_b; auto. eapply H3, H9.
             intros; eapply remove_in_b', H8; auto.      
        }
        assert (ext G).
        {
           intros ? ? ? ?; subst G.
           simpl. erewrite H9, H5; [reflexivity|eassumption].
        }
        assert (in_b a l).
        {
           specialize (iter_bound _ _ H3 p); intro.
           exploit H10; [intros ? H'; inversion H'|].
           apply H3 in E'; auto. intros.
           apply H10 in H11; auto.
           eapply H3.
           + apply (H10 (S n)).
           + auto.
        }
        assert (length (remove deceq a l) < S n).
        {
           rewrite Heqn; apply remove_in_b_length; auto.
        }
        assert (forall z, p z -> in_b z (remove deceq a l)).
        {
           intros.
           destruct (deceq a z); subst.
           + exfalso.
             apply (infl_iter _ H2 ((S n))) in H12.
             rewrite H12 in E; inversion E.
           + eapply remove_in_b; eauto.
        }
        specialize (H0 _ H11 _ eq_refl _ H12 H7 H8 H6 H9).
        assert (forall k, k <= S n -> iter F k p a = false).
        {
           intros; case_eq (iter F k p a); intros; auto.
           replace (S n) with ((S n - k) + k) in E; [|lia].
           rewrite iter_plus in E.
           apply (infl_iter _ H2 (S n - k)) in H14.
           rewrite H14 in E; assumption.
        }
        assert (forall k, k <= S n -> iter F k p == iter G k p).
        {  
          induction k; simpl; intros; [reflexivity|].
          erewrite H9; [|intro; symmetry; eapply IHk; lia].
          subst G; simpl.
          destruct (deceq a x); subst.
          + rewrite H13;[|lia].
            specialize (H13 (S k)).
            simpl in *; rewrite H13; auto.
          + reflexivity.
        }
        assert (H15 := H14 (S (length (remove deceq a l)))); exploit H15;[lia|].
        simpl in H15.
        assert (F (@iter (Σ -> bool) F (@length Σ (@remove Σ deceq a l)) p) ==
       @iter (Σ -> bool) F (@length Σ (@remove Σ deceq a l)) p).
        {
          intros ?; rewrite H15, H14;[|lia]; auto.
        }
        clear H15 H14 H13 H12 H10 H9 H8 H7 H6 H3 H2 H1 H4.
        clear G H0 H.
        remember (length (remove deceq a l)) as k.
        clear Heqk Heqn l.
        assert (forall m, iter F (m + k) p == iter F k p).
        {
           induction m; simpl; auto.
           intro; erewrite H5; [|eapply IHm]; apply H16.
        }
        replace (S n) with ((S n - k) + k) in E;[|lia].
        rewrite H in E.
        rewrite <-(H (S ((S n) - k))) in E.
        replace (S (S n - k) + k) with (S (S n)) in E;[|lia].
        simpl in *. rewrite E' in E.
        inversion E.
      + apply H2 in E.
        rewrite E in E'; inversion E'.
   Qed.

  Lemma In_in_b : forall x l, In x l -> in_b x l.
    induction l; simpl.
    + intros H; inversion H.
    + intros [|]; destruct (deceq x a); auto; try reflexivity.
      exfalso; auto.
  Qed.

  End Kleenefin.
End Kleenefin.

Theorem Kleene : forall {Σ : Alphabet} (F : (Σ -> bool) -> (Σ -> bool)) (p : Σ -> bool), Kleenefin.mono _ F -> Kleenefin.ext _ F -> Kleenefin.inflationary _ F -> {n : nat | F (iter F n p) == iter F n p}.
   intros.
   refine (exist _ _ (Kleenefin.Kleene _ (FT_deceq _) (FT_enum Σ) _ _ _ H1 _ _ _)); eauto.
   + intros ? ?.
     apply Kleenefin.In_in_b, FT_enum_all.
   + intros ? _ ? ?; apply Kleenefin.In_in_b, FT_enum_all.
Defined.

Lemma exf_mono : forall {Σ : Alphabet} (p q : Σ -> bool), (forall x, p x -> q x) -> exf p -> exf q.
  intros.
  apply exf_exists in H0; apply exf_exists; firstorder.
Qed.

Lemma exf_ext : forall {Σ : Alphabet} (p q : Σ -> bool), p == q -> exf p = exf q.
  intros.
  case_eq (exf p); case_eq (exf q); intros; try reflexivity; exfalso.
  + apply exf_exists in H1.
    assert (exists x, q x) by (destruct H1 as [w ?]; exists w; rewrite H in *; eauto).
    apply exf_exists in H2; congruence.
  + apply exf_exists in H0.
    assert (exists x, p x) by (destruct H0 as [w ?]; exists w; rewrite <-H in *; eauto).
    apply exf_exists in H2; congruence.
Qed.

Lemma andb_mono : forall a b c d : bool, (a -> b) -> (c -> d) -> (a && c) -> b && d.
  intros [|] [|] [|] [|]; auto.
Qed.

Lemma orb_mono : forall a b c d : bool, (a -> b) -> (c -> d) -> (a || c) -> b || d.
  intros [|] [|] [|] [|]; auto.
Qed.
